__author__ = 'Ahmet Bakan, Anindita Dutta, Wenzhi Mao, James Krieger'

from prody.tests import TestCase

from numpy import array, log, zeros, char, ones, fromfile
from numpy.testing import assert_array_equal, assert_array_almost_equal

from prody.tests.datafiles import *

from prody import LOGGER, calcShannonEntropy, buildMutinfoMatrix, parseMSA, parsePDB
from prody import calcMSAOccupancy, buildSeqidMatrix, uniqueSequences
from prody import buildOMESMatrix, buildSCAMatrix, calcMeff, buildMSA
from prody import buildDirectInfoMatrix

LOGGER.verbosity = None

FASTA = parseMSA(pathDatafile('msa_Cys_knot.fasta'))
FASTA_ALPHA = char.isalpha(FASTA._msa)
FASTA_UPPER = char.upper(FASTA._msa)

FASTA_NUMBER, FASTA_LENGTH = FASTA_ALPHA.shape
FASTA_EYE = zeros((FASTA_NUMBER, FASTA_NUMBER))
for i in range(FASTA_NUMBER):
    FASTA_EYE[i, i] = 1
    for j in range(i + 1, FASTA_NUMBER):
        score = 0.0
        ncols = 0
        for k in range(FASTA_LENGTH):
            if FASTA_ALPHA[i, k] or FASTA_ALPHA[j, k]:
                if FASTA_UPPER[i, k] == FASTA_UPPER[j, k]:
                    score += 1
                ncols += 1
        FASTA_EYE[i, j] = FASTA_EYE[j, i] = score / ncols

ags = parsePDB([DATA_FILES['3hsy']['path'], 
                DATA_FILES['3o21']['path']])

class TestCalcShannonEntropy(TestCase):

    def testSixSequences(self):

        msa = array([list('AAAAaaaaAAAAaaaa'),
                     list('AAACaaacAAACaaac'),
                     list('AACDaacdAACDaacd'),
                     list('ACCEacceacceACCE'),
                     list('ACDFacdfacdfACDF'),
                     list('ACDGacdgacdgACDG')], dtype='|S1')

        expect = -log(1. / array([1, 2, 3, 6] * 4))
        result = calcShannonEntropy(msa)
        assert_array_almost_equal(expect, result)

    def testTwenty(self):

        msa = array([[char] for char in 'ACDEFGHIKLMNPQRSTVWY'], dtype='|S1')

        expect = -log(1. / 20)
        result = calcShannonEntropy(msa)
        assert_array_almost_equal(expect, result)

    def testSmallProbability(self):

        msa = zeros((1000000, 1), '|S1')
        msa[0] = 'A'
        msa[1:] = 'C'
        expect = array([1., 999999.]) / 1000000
        expect = - (expect * log(expect)).sum()
        result = calcShannonEntropy(msa)
        assert_array_almost_equal(expect, result)

    def testAmbiguous(self):

        msa = array([list('bjzxBJZX'),
                     list('bjzxBJZX'), ], dtype='|S1')

        expect = -log(1. / array([2, 2, 2, 20] * 2))
        result = calcShannonEntropy(msa)
        assert_array_almost_equal(expect, result)

    def testGapDividend(self):

        msa = array([list('AAAA'),
                     list('AAAC'),
                     list('AACD'),
                     list('ACCE'),
                     list('ACDF'),
                     list('ACDG'),
                     list('----')], dtype='|S1')

        expect = -log(1. / array([1, 2, 3, 6]))
        result = calcShannonEntropy(msa, omitgaps=True)
        assert_array_almost_equal(expect, result)

"""
    def testSixSequences3(self):

        msa = array([list('AAAA'),
                     list('AAAB'),
                     list('AABC'),
                     list('ABBD'),
                     list('ABCE'),
                     list('ABCF')])

        expect = -log(1. / array([1, 2, 3, 6]))
        result = calcInfoEntropy(msa)
        assert_array_almost_equal(expect, result)
"""


class TestCalcMutualInfo(TestCase):

    def testSixSequences(self):

        msa = array([list('ACCA'),
                     list('ACDA'),
                     list('ACEC'),
                     list('ACGC')], dtype='|S1')

        expect = array([[0., 0., 0., 0.],
                        [0., 0., 0., 0.],
                        [0., 0., 0., log(2.)],
                        [0., 0., log(2.), 0.], ])
        result = buildMutinfoMatrix(msa)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildMutinfoMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testTwenty(self):

        seq = 'ACDEFGHIKLMNPQRSTVWY'
        msa = array([[s, s] for s in seq], dtype='|S1')

        expect = log(20.)
        expect = array([[0., expect],
                        [expect, 0.]])
        result = buildMutinfoMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildMutinfoMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testTwentyReversed(self):

        seq = 'ACDEFGHIKLMNPQRSTVWY'
        msa = array([[s, seq[-i-1]] for i, s in enumerate(seq)], dtype='|S1')

        expect = log(20.)
        expect = array([[0., expect],
                        [expect, 0.]])
        result = buildMutinfoMatrix(msa)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildMutinfoMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testAmbiguity(self):

        msa = array([list('OX'),
                     list('XO')], dtype='|S1')

        expect = array([[0., log(2.)],
                        [log(2.), 0.]])
        result = buildMutinfoMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildMutinfoMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testNoAmbiguity(self):

        msa = array([list('OX'),
                     list('XO')], dtype='|S1')

        expect = array([[0., log(2.)],
                        [log(2.), 0.]])
        result = buildMutinfoMatrix(msa, ambiquity=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildMutinfoMatrix(msa, ambiquity=False, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testAmbiguity2(self):

        msa = array([list('AB'),
                     list('BZ')], dtype='|S1')
        expect = (2 * .25 * log(.25 / .5 / .25) +
                  4 * .125 * log(.125 / .25 / .25))
        expect = array([[0., expect],
                        [expect, 0.]])
        result = buildMutinfoMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildMutinfoMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testAmbiguity3(self):

        msa = array([list('XX')], dtype='|S1')

        expect = zeros((2, 2))
        result = buildMutinfoMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildMutinfoMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testAmbiguity4(self):

        msa = array([list('Bb'),
                     list('jJ'),
                     list('Zz'), ], dtype='|S1')

        expect = log((1./12) / (1./6) / (1./6))
        expect = array([[0., expect],
                        [expect, 0.]])
        result = buildMutinfoMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildMutinfoMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testAmbiguity5(self):

        expect = array([[0., 0.],
                        [0., 0.]])

        for seq in ['bx', 'Xb', 'jX', 'Xj', 'xz', 'ZX',
                    'bj', 'jb', 'bz', 'zb', 'jz', 'zj']:
            msa = array([list(seq)], dtype='|S1')
            result = buildMutinfoMatrix(msa, debug=False)
            assert_array_almost_equal(expect, result, err_msg=seq + ' failed')

    def testAmbiguity6(self):

        expect = zeros((2, 2))

        for seq in ['bb', 'jj', 'zz']:
            msa = array([list(seq)], dtype='|S1')
            result = buildMutinfoMatrix(msa, debug=False)
            assert_array_almost_equal(expect, result, err_msg=seq + ' failed')

    def testAmbiguity7(self):

        msa = array([list('bx'),
                     list('xb')], dtype='|S1')
        expect = (72 * 0.0125 * log(0.0125/0.0250/0.275) +
                  4 * 0.0250 * log(0.0250/0.275/0.275))
        expect = array([[0., expect],
                        [expect, 0.]])
        result = buildMutinfoMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildMutinfoMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testInf(self):

        msa = zeros((500, 10), '|S1')
        msa.fill('.')
        msa[95, 8] = 's'
        msa[95, 9] = 'i'
        expect = zeros((10, 10))
        expect[8, 9] = expect[9, 8] = 0.002 * log(500.) + .998 * log(1. / .998)
        result = buildMutinfoMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildMutinfoMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testNorm(self):

        seq = 'ACDEFGHIKLMNPQRSTVWY'
        msa = array([[s, seq[-i-1]] for i, s in enumerate(seq)], dtype='|S1')

        expect = 1.
        expect = array([[0., expect],
                        [expect, 0.]])
        result = buildMutinfoMatrix(msa, norm=True)
        assert_array_almost_equal(expect, result, err_msg='norm failed')

    def testNorm2(self):

        seq = 'ACDEFGHIKLMNPQRSTVWY'
        msa = array([[s, 'O' if i % 2 else 'U'] for i, s in enumerate(seq)],
                    dtype='|S1')

        expect = log(1./20. / (1./20. * 1./2.)) / (-log(1./20.))
        expect = array([[0., expect],
                        [expect, 0.]])
        result = buildMutinfoMatrix(msa, norm=True)
        assert_array_almost_equal(expect, result, err_msg='norm failed')


class TestCalcMSAOccupancy(TestCase):

    def testResidueCount(self):

        assert_array_equal(calcMSAOccupancy(FASTA, 'residue', count=1),
                           FASTA_ALPHA.sum(0))

    def testSequenceCount(self):

        assert_array_equal(calcMSAOccupancy(FASTA, 'sequence', count=1),
                           FASTA_ALPHA.sum(1))

    def testResidueOccupancy(self):

        assert_array_equal(calcMSAOccupancy(FASTA, 'residue'),
                           FASTA_ALPHA.sum(0) / (FASTA.numSequences() * 1.0))

    def testSequenceOccupancy(self):

        assert_array_equal(calcMSAOccupancy(FASTA, 'sequence'),
                           FASTA_ALPHA.sum(1) / (FASTA.numResidues() * 1.0))


class TestIdentity(TestCase):

    def testIdentityMatrix(self):

        assert_array_almost_equal(FASTA_EYE, buildSeqidMatrix(FASTA))

    def testIdentityMatrixNonTurbo(self):

        assert_array_almost_equal(FASTA_EYE,
                                  buildSeqidMatrix(FASTA, turbo=False))


class TestUnique(TestCase):

    def testUnique(self):

        seqid = 0.98
        unique = ones(FASTA_NUMBER, bool)
        for i in range(FASTA_NUMBER):
            if not unique[i]:
                continue
            for j in range(i+1, FASTA_NUMBER):
                if FASTA_EYE[i, j] >= seqid:
                    unique[j] = False

        assert_array_equal(unique, uniqueSequences(FASTA, seqid))

    def testUnique2(self):

        seqid = 0.5
        unique = ones(FASTA_NUMBER, bool)
        for i in range(FASTA_NUMBER):
            if not unique[i]:
                continue
            for j in range(i+1, FASTA_NUMBER):
                if FASTA_EYE[i, j] >= seqid:
                    unique[j] = False

        assert_array_equal(unique, uniqueSequences(FASTA, seqid))


class TestCalcOMES(TestCase):

    def testZero(self):

        msa = array([list('ACCA'),
                     list('ACDA'),
                     list('ACCC'),
                     list('ACDC')], dtype='|S1')

        expect = array([[0., 0., 0., 0.],
                        [0., 0., 0., 0.],
                        [0., 0., 0., 0.],
                        [0., 0., 0., 0.], ])
        result = buildOMESMatrix(msa)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildOMESMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testFourSequences(self):

        msa = array([list('ACCA'),
                     list('ACDA'),
                     list('ACDC'),
                     list('ACDC')], dtype='|S1')

        expect = array([[0., 0., 0., 0.],
                        [0., 0., 0., 0.],
                        [0., 0., 0., 4./3],
                        [0., 0., 4./3, 0.], ])
        result = buildOMESMatrix(msa)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildOMESMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testTwenty(self):

        seq = 'ACDEFGHIKLMNPQRSTVWY'
        msa = array([[s, s] for s in seq], dtype='|S1')

        expect = array([[0., 380.],
                        [380., 0.]])
        result = buildOMESMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildOMESMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testTwentyReversed(self):

        seq = 'ACDEFGHIKLMNPQRSTVWY'
        msa = array([[s, seq[-i-1]] for i, s in enumerate(seq)], dtype='|S1')

        expect = array([[0., 380.],
                        [380., 0.]])
        result = buildOMESMatrix(msa)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildOMESMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testAmbiguity(self):

        msa = array([list('OX'),
                     list('XO')], dtype='|S1')

        expect = array([[0., 2.],
                        [2., 0.]])
        result = buildOMESMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildOMESMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testNoAmbiguity(self):

        msa = array([list('OX'),
                     list('XO')], dtype='|S1')

        expect = array([[0., 2.],
                        [2., 0.]])
        result = buildOMESMatrix(msa, ambiquity=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildOMESMatrix(msa, ambiquity=False, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testAmbiguity2(self):

        msa = array([list('AB'),
                     list('BZ')], dtype='|S1')
        expect = array([[0., 2.],
                        [2., 0.]])
        result = buildOMESMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildOMESMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testAmbiguity3(self):

        msa = array([list('XX')], dtype='|S1')

        expect = zeros((2, 2))
        result = buildOMESMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildOMESMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testAmbiguity4(self):

        msa = array([list('Bb'),
                     list('jJ'),
                     list('Zz'), ], dtype='|S1')

        expect = array([[0., 6.],
                        [6., 0.]])
        result = buildOMESMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildOMESMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testAmbiguity5(self):

        expect = array([[0., 0.],
                        [0., 0.]])

        for seq in ['bx', 'Xb', 'jX', 'Xj', 'xz', 'ZX',
                    'bj', 'jb', 'bz', 'zb', 'jz', 'zj']:
            msa = array([list(seq)], dtype='|S1')
            result = buildOMESMatrix(msa, debug=False)
            assert_array_almost_equal(expect, result, err_msg=seq + ' failed')

    def testAmbiguity6(self):

        expect = zeros((2, 2))

        for seq in ['bb', 'jj', 'zz']:
            msa = array([list(seq)], dtype='|S1')
            result = buildOMESMatrix(msa, debug=False)
            assert_array_almost_equal(expect, result, err_msg=seq + ' failed')

    def testAmbiguity7(self):

        msa = array([list('bx'),
                     list('xb')], dtype='|S1')
        expect = array([[0., 162./121],
                        [162./121, 0.]])
        result = buildOMESMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildOMESMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testInf(self):

        msa = zeros((500, 10), '|S1')
        msa.fill('.')
        msa[95, 8] = 's'
        msa[95, 9] = 'i'
        expect = zeros((10, 10))
        expect[8, 9] = expect[9, 8] = 500.
        result = buildOMESMatrix(msa, debug=False)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildOMESMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')


class TestCalcSCA(TestCase):

    def testZero(self):

        msa = array([list('ACCD'),
                     list('ACDD'),
                     list('ACCC'),
                     list('ACDC')], dtype='|S1')

        expect = array([log(0.975/.025)*.5, log(0.95/.05)*.5])
        weight = ((expect ** 2).sum())**.5
        expect = expect / weight * array([log(0.975/.025), log(0.95/.05)])
        expect = (expect ** 2).mean() - (expect.mean()) ** 2
        expect = array([[0., 0., 0., 0.],
                        [0., 0., 0., 0.],
                        [0., 0., expect, 0.],
                        [0., 0., 0., expect], ])
        result = buildSCAMatrix(msa)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildSCAMatrix(msa, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')

    def testMATLAB(self):

        sca = fromfile(pathDatafile('msa_Cys_knot_sca.dat'))
        expect = sca.reshape((10, 10))
        fasta = FASTA[:, :10]
        result = buildSCAMatrix(fasta, turbo=True)
        assert_array_almost_equal(expect, result, err_msg='turbo failed')
        result = buildSCAMatrix(fasta, turbo=False)
        assert_array_almost_equal(expect, result, err_msg='w/out turbo failed')


class TestCalcMeff(TestCase):

    def testZero1(self):

        msa = array([list('ACCD')] * 100, dtype='|S1')
        expect = 1.
        result = calcMeff(msa)
        assert_array_almost_equal(expect, result)

        result = calcMeff(msa, seqid=0.9)
        assert_array_almost_equal(expect, result)

        result = calcMeff(msa, seqid=0.4)
        assert_array_almost_equal(expect, result)

        expect = (1., zeros((100)) + 1./100)
        result = calcMeff(msa, weight=True)
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

        result = calcMeff(msa, seqid=0.9, weight=True)
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

        result = calcMeff(msa, seqid=0.4, weight=True)
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

    def testZero2(self):

        msa = array([list('AACC')] * 50 + [list('AADD')] * 50, dtype='|S1')
        expect = 2.
        result = calcMeff(msa)
        assert_array_almost_equal(expect, result)

        result = calcMeff(msa, seqid=0.9)
        assert_array_almost_equal(expect, result)

        expect = 1.
        result = calcMeff(msa, seqid=0.4)
        assert_array_almost_equal(expect, result)

        expect = (2., zeros((100)) + 1./50)
        result = calcMeff(msa, weight=True)
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

        result = calcMeff(msa, seqid=0.9, weight=True)
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

        expect = (1., zeros((100)) + 1./100)
        result = calcMeff(msa, seqid=0.4, weight=True)
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

    def testTwenty(self):

        seq = 'ACDEFGHIKLMNPQRSTVWY'
        msa = array([[s, s] for s in seq], dtype='|S1')
        expect = 20.
        result = calcMeff(msa)
        assert_array_almost_equal(expect, result)

        result = calcMeff(msa, seqid=0.9)
        assert_array_almost_equal(expect, result)

        result = calcMeff(msa, seqid=0.4)
        assert_array_almost_equal(expect, result)

        expect = (20., ones(20))
        result = calcMeff(msa, weight=True)
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

        result = calcMeff(msa, seqid=0.9, weight=True)
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

        result = calcMeff(msa, seqid=0.4, weight=True)
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

    def testTwentyReversed(self):

        seq = 'ACDEFGHIKLMNPQRSTVWY'
        msa = array([[s, seq[-i-1]] for i, s in enumerate(seq)], dtype='|S1')
        expect = 20.
        result = calcMeff(msa)
        assert_array_almost_equal(expect, result)

        result = calcMeff(msa, seqid=0.9)
        assert_array_almost_equal(expect, result)

        result = calcMeff(msa, seqid=0.4)
        assert_array_almost_equal(expect, result)

        expect = (20., ones(20))
        result = calcMeff(msa, weight=True)
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

        result = calcMeff(msa, seqid=0.9, weight=True)
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

        result = calcMeff(msa, seqid=0.4, weight=True)
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

    def testMATLAB(self):

        expect = 1.8416666666666664e+01
        result = calcMeff(FASTA, refine=True)
        assert_array_almost_equal(expect, result)
        result = calcMeff(FASTA, refine=True, weight=True)
        expect = (expect,
                  array([1./3, 1./3, 1./4, 1./2, 1., 1., 1., 1., 1./3, 1./3,
                         1./3, 1./2, 1./2, 1., 1., 1., 1., 1., 1., 1., 1., 1.,
                         1., 1./2, 1./2], dtype='float'))
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

        expect = 23
        result = calcMeff(FASTA, seqid=0.9, refine=True)
        assert_array_almost_equal(expect, result)
        result = calcMeff(FASTA, seqid=0.9, refine=True, weight=True)
        expect = (expect,
                  array([0.5, 0.5, 1., 1., 1., 1., 1., 1., 1., 1., 1., 0.5,
                         0.5, 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.,
                         1.], dtype='float'))
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')

        expect = 4.66689144189144
        result = calcMeff(FASTA, seqid=0.4, refine=True)
        assert_array_almost_equal(expect, result)
        result = calcMeff(FASTA, seqid=0.4, refine=True, weight=True)
        expect = (expect,
                  array([0.071428571, 0.071428571, 0.066666667, 0.083333333,
                         0.076923077, 0.066666667, 0.076923077, 0.090909091,
                         0.090909091, 0.071428571, 0.076923077, 0.076923077,
                         0.1, 0.2, 0.166666667, 0.25, 1., 0.25, 0.25, 0.25,
                         0.25, 0.333333333, 0.5, 0.071428571,
                         0.125, ], dtype='float'))
        assert_array_almost_equal(expect[0], result[0],
                                  err_msg='weight failed')
        assert_array_almost_equal(expect[1], result[1],
                                  err_msg='weight failed')


class TestDirectInfo(TestCase):

    def testZero(self):

        msa = array([list('ACCY')] * 100, dtype='|S1')
        expect = array([[0., 0.66325166608, 0.66325166608, 0.66222154839],
                        [0.66325166608, 0., 0.66325166608, 0.66222154839],
                        [0.66325166608, 0.66325166608, 0., 0.66222154839],
                        [0.66222154839, 0.66222154839, 0.66222154839, 0.], ])
        result = buildDirectInfoMatrix(msa)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.3
        expect = array([[0.,  0.53477644,  0.53477644,  0.58513139],
                        [0.53477644,  0.,  0.53477644,  0.58513139],
                        [0.53477644,  0.53477644,  0.,  0.58513139],
                        [0.58513139,  0.58513139,  0.58513139,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.8
        expect = array([[0.,  0.41397564,  0.41397564,  0.41295744],
                        [0.41397564,  0.,  0.41397564,  0.41295744],
                        [0.41397564,  0.41397564,  0.,  0.41295744],
                        [0.41295744,  0.41295744,  0.41295744,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')

    def testZero1(self):

        msa = array([list('ACCD')] * 100, dtype='|S1')
        expect = array([[0.,  0.13010138,  0.13010138,  0.13008827],
                        [0.13010138,  0.,  0.13010138,  0.13008827],
                        [0.13010138,  0.13010138,  0.,  0.13008827],
                        [0.13008827,  0.13008827,  0.13008827,  0.]])
        result = buildDirectInfoMatrix(msa)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.3
        expect = array([[0.,  0.23225802,  0.23225802,  0.23225606],
                        [0.23225802,  0.,  0.23225802,  0.23225606],
                        [0.23225802,  0.23225802,  0.,  0.23225606],
                        [0.23225606,  0.23225606,  0.23225606,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.8
        expect = array([[0.,  0.03910082,  0.03910082,  0.03910178],
                        [0.03910082,  0.,  0.03910082,  0.03910178],
                        [0.03910082,  0.03910082,  0.,  0.03910178],
                        [0.03910178,  0.03910178,  0.03910178,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')

    def testZero2(self):

        msa = array([list('AAYY')] * 50 + [list('YYDD')] * 50, dtype='|S1')
        expect = array([[0., 1.0248086877, 1.0001784999, 1.0001784999],
                        [1.0248086877, 0., 1.0001784999, 1.0001784999],
                        [1.0001784999, 1.0001784999, 0., 1.0248086877],
                        [1.0001784999, 1.0001784999, 1.0248086877, 0.], ])
        result = buildDirectInfoMatrix(msa)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.3
        expect = array([[0.,  1.30637709,  1.00041168,  1.00041168],
                       [1.30637709,  0.,  1.00041168,  1.00041168],
                       [1.00041168,  1.00041168,  0.,  1.30637709],
                       [1.00041168,  1.00041168,  1.30637709,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.8
        expect = array([[0.,  0.55686383,  0.55605426,  0.55605426],
                       [0.55686383,  0.,  0.55605426,  0.55605426],
                       [0.55605426,  0.55605426,  0.,  0.55686383],
                       [0.55605426,  0.55605426,  0.55686383,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')

    def testZero3(self):

        msa = array([list('AACC')] * 50 + [list('CCDD')] * 50, dtype='|S1')
        expect = array([[0.,  0.23179074,  0.23178758,  0.23178758],
                        [0.23179074,  0.,  0.23178758,  0.23178758],
                        [0.23178758,  0.23178758,  0.,  0.23178758],
                        [0.23178758,  0.23178758,  0.23178758,  0.]])
        result = buildDirectInfoMatrix(msa)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.3
        expect = array([[0.,  0.54966352,  0.54965557,  0.54965557],
                       [0.54966352,  0.,  0.54965557,  0.54965557],
                       [0.54965557,  0.54965557,  0.,  0.54965565],
                       [0.54965557,  0.54965557,  0.54965565,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.8
        expect = array([[0.,  0.04349673,  0.0434971,  0.0434971],
                        [0.04349673,  0.,  0.0434971,  0.0434971],
                        [0.0434971,  0.0434971,  0.,  0.04349711],
                        [0.0434971,  0.0434971,  0.04349711,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')

    def testZero4(self):

        msa = array([list('AAYY')] * 50 + [list('AADD')] * 50, dtype='|S1')
        expect = array([[0.,  0.68512031,  0.54656421,  0.54656421],
                        [0.68512031,  0.,  0.54656421,  0.54656421],
                        [0.54656421,  0.54656421,  0.,  1.05788193],
                        [0.54656421,  0.54656421,  1.05788193,  0.]])
        result = buildDirectInfoMatrix(msa)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.3
        expect = array([[0.,  0.65929972,  0.52603587,  0.52603587],
                       [0.65929972,  0.,  0.52603587,  0.52603587],
                       [0.52603587,  0.52603587,  0.,  1.99332504],
                       [0.52603587,  0.52603587,  1.99332504,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.8
        expect = array([[0.,  0.48628027,  0.27348214,  0.27348214],
                       [0.48628027,  0.,  0.27348214,  0.27348214],
                       [0.27348214,  0.27348214,  0.,  0.65235465],
                       [0.27348214,  0.27348214,  0.65235465,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')

    def testTwenty(self):

        seq = 'ACDEFGHIKLMNPQRSTVWY'
        msa = array([[s, s] for s in seq], dtype='|S1')
        expect = array([[0., 3.0302471958885744],
                        [3.0302471958885744, 0.]])
        result = buildDirectInfoMatrix(msa)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.3
        expect = array([[0.,  3.0757238],
                        [3.0757238,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.8
        expect = array([[0.,  1.89488914],
                        [1.89488914,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')

    def testTwentyReversed(self):

        seq = 'ACDEFGHIKLMNPQRSTVWY'
        msa = array([[s, seq[-i-1]] for i, s in enumerate(seq)], dtype='|S1')
        expect = array([[0., 3.030670764986982],
                        [3.030670764986982, 0.]])
        result = buildDirectInfoMatrix(msa)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.3
        expect = array([[0.,  3.12974624],
                        [3.12974624,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.3)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.3, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        # pseudo_weight=0.8
        expect = array([[0.,  1.89488914],
                        [1.89488914,  0.]])
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(msa, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.9, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.9, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')
        result = buildDirectInfoMatrix(msa, seqid=0.4, pseudo_weight=0.8)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(
            msa, seqid=0.4, pseudo_weight=0.8, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')

    def testMATLAB8(self):

        di = fromfile(pathDatafile('msa_Cys_knot_di.dat'))
        expect = di.reshape((8, 8))
        fasta = FASTA[:, :8]
        result = buildDirectInfoMatrix(fasta)
        assert_array_almost_equal(
            expect, result, err_msg='w/out refine failed')
        result = buildDirectInfoMatrix(fasta, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')

    def testMATLAB10(self):

        di = fromfile(pathDatafile('msa_Cys_knot_di.dat'))
        expect = di.reshape((8, 8))
        fasta = FASTA[:, :10]
        result = buildDirectInfoMatrix(fasta, refine=True)
        assert_array_almost_equal(expect, result, err_msg='refine failed')

class TestBuildMSA(TestCase):

    def testBuildMSAlocal(self):
        sequences = [ags[0].protein["A"].getSequence(),
                     ags[1].protein["A"].getSequence()]

        expect1 = parseMSA(pathDatafile('msa_3hsyA_3o21A.fasta'))
        expect2 = parseMSA(pathDatafile('msa_3hsyA_3o21A_new.fasta'))
        result = buildMSA(sequences, method="local", labels=["A2", "A3"])
        assert result in (expect1, expect2), "The list of expected buildMSA results did not contain " + result

    def testBuildMSAclustalw(self):
        sequences = [ags[0].protein["A"].getSequence(),
                     ags[1].protein["A"].getSequence()]

        expect1 = parseMSA(pathDatafile('msa_3hsyA_3o21A_clustalw.fasta'))
        result = buildMSA(sequences, method="clustalw", labels=["A2", "A3"])
        assert result == expect1, "The list of expected buildMSA clustalw results did not contain " + result



# -*- coding: utf-8 -*-
"""This module defines :class:`Bond` for dealing with bond information provided
by using the :meth:`.AtomGroup.setBonds` or :meth:`.AtomGroup.inferBonds` method."""

from numbers import Integral
import numpy as np

__all__ = ['Bond']

class Bond(object):

    """A pointer class for bonded atoms.  Following built-in functions are
    customized for this class:

    * :func:`len` returns bond length, i.e. :meth:`getLength`
    * :func:`iter` yields :class:`~.Atom` instances"""

    __slots__ = ['_ag', '_acsi', '_indices', '_bondOrder']
    _bondType = {1:'single', 2:'double', 3:'triple', 4:'aromatic', 5:'amide', }

    def __init__(self, ag, indices, acsi=None):

        self._ag = ag
        i, j  = self._indices = np.array(indices)
        if self._ag._bondOrders is not None:
            if i>j:
                a=i
                i=j
                j=a
            self._bondOrder = self._ag._bondOrders[self._ag._bondIndex['%d %d'%(i,j)]]
        else:
            self._bondOrder = 1 # single bond by default
        if acsi is None:
            self._acsi = ag.getACSIndex()
        else:
            self._acsi = acsi

    def __repr__(self):
        one, two = self._indices
        names = self._ag._getNames()
        return '<Bond: {0}({1})--{2}({3}) from {4}, bond order: {5}>'.format(
            names[one], one, names[two], two, str(self._ag),
            self._bondType[self._bondOrder])

    def __str__(self):

        one, two = self._indices
        names = self._ag._getNames()
        return '{0}({1})--{2}({3})'.format(
                                            names[one], one, names[two], two)

    def __eq__(self, other):

        return (isinstance(other, Bond) and other.getAtomGroup() is self._ag
                and (np.all(other.getIndices() == self._indices) or
                 np.all(other.getIndices() == list(reversed(self._indices)))))

    def __ne__(self, other):

        return not self.__eq__(other)

    def __len__(self):

        return self.getLength()

    def __iter__(self):

        for index in self._indices:
            yield self._ag[index]

    def getAtomGroup(self):
        """Returns atom group."""

        return self._ag

    def getAtoms(self):
        """Returns bonded atoms."""

        return (self._ag[self._indices[0]], self._ag[self._indices[1]])

    def getIndices(self):
        """Returns indices of bonded atoms."""

        return self._indices.copy()

    def getLength(self):
        """Returns bond length."""

        vector = self.getVector()
        return np.multiply(vector, vector, vector).sum() ** 0.5

    def getVector(self):
        """Returns bond vector that originates from the first atom."""

        one, two = self._indices
        acsi = self.getACSIndex()
        return self._ag._coords[acsi, two] - self._ag._coords[acsi, one]

    def getACSIndex(self):
        """Returns index of the coordinate set."""

        acsi = self._acsi
        if acsi >= self._ag._n_csets:
            raise ValueError('{0} has fewer coordsets than assumed by {1}'
                             .format(str(self._ag), str(self)))
        return acsi

    def setACSIndex(self, index):
        """Set the coordinate set at *index* active."""

        if self._ag._coords is None:
            raise AttributeError('coordinates are not set')

        if not isinstance(index, Integral):
            raise TypeError('index must be an integer')

        n_csets = self._ag._n_csets
        if n_csets <= index or n_csets < abs(index):
            raise IndexError('coordinate set index is out of range')

        if index < 0:
            index += n_csets

        self._acsi = index


def evalBonds(bonds, n_atoms):
    """Returns an array mapping atoms to their bonded neighbors and an array
    that stores number of bonds made by each atom."""

    numbonds = np.bincount(bonds.reshape((bonds.shape[0] * 2)))
    bmap = np.zeros((n_atoms, numbonds.max()), int)
    bmap.fill(-1)
    index = np.zeros(n_atoms, int)
    for bond in bonds:
        a, b = bond
        bmap[a, index[a]] = b
        bmap[b, index[b]] = a
        index[bond] += 1
    return bmap, numbonds


def trimBonds(bonds, indices):
    """Returns bonds between atoms at given indices."""

    iset = set(indices)
    bonds = [bond for bond in bonds if bond[0] in iset and bond[1] in iset]
    if bonds:
        newindices = np.zeros(indices.max()+1, int)
        newindices[indices] = np.arange(len(indices))
        return newindices[np.array(bonds)]



# -*- coding: utf-8 -*-
"""This module defines functions for handling PDB sequence clusters."""

import os.path

from numpy import array, abs

from prody import LOGGER, SETTINGS, getPackagePath
from prody.utilities import openFile, openURL, pystr, isListLike

from numbers import Integral

__all__ = ['fetchPDBClusters', 'loadPDBClusters', 'listPDBCluster']

PDB_CLUSTERS = {30: None, 40: None, 50: None, 70: None,
                90: None, 95: None, 100: None}
PDB_CLUSTERS_UPDATE_WARNING = True
PDB_CLUSTERS_SQIDS = array(list(PDB_CLUSTERS))
PDB_CLUSTERS_SQIDS.sort()
PDB_CLUSTERS_SQID_STR = ', '.join([str(key) for key in PDB_CLUSTERS_SQIDS])

def loadPDBClusters(sqid=None):
    """Load previously fetched PDB sequence clusters from disk to memory."""

    PDB_CLUSTERS_PATH = os.path.join(getPackagePath(), 'pdbclusters')
    if sqid is None:
        sqid_list = list(PDB_CLUSTERS)
        LOGGER.info('Loading all PDB sequence clusters.')
    else:
        assert isinstance(sqid, Integral), 'sqid must be an integer'
        if sqid not in PDB_CLUSTERS:
            raise ValueError('PDB cluster data is not available for sequence '
                             'identity {0}%, try one of {1}'
                             .format(sqid, PDB_CLUSTERS_SQID_STR))
        LOGGER.info('Loading PDB sequence clusters for sequence identity '
                    '{0}.'.format(sqid))
        sqid_list = [sqid]
    global PDB_CLUSTERS_UPDATE_WARNING
    for sqid in sqid_list:
        filename = os.path.join(PDB_CLUSTERS_PATH,
                                'bc-{0}.out.gz'.format(sqid))
        if not os.path.isfile(filename):
            fetchPDBClusters(sqid)

        if PDB_CLUSTERS_UPDATE_WARNING:
            import time
            diff = (time.time() - os.path.getmtime(filename)) / 604800.
            if diff > 1.:
                LOGGER.warning('PDB sequence clusters are {0:.1f} week(s) old,'
                               ' call `fetchPDBClusters` to receive updates.'
                               .format(diff))
                PDB_CLUSTERS_UPDATE_WARNING = False
        inp = openFile(filename)
        clusters_str = pystr(inp.read())

        clusters = []
        for cluster_str in clusters_str.split('\n'):
            cluster_str = cluster_str.strip()
            if len(cluster_str):
                cluster = [tuple(item.split('_')) for item in cluster_str.split()]
                clusters.append(cluster)

        PDB_CLUSTERS[sqid] = clusters
        inp.close()

    if sqid is None:
        return PDB_CLUSTERS
    else:
        return clusters


def listPDBCluster(pdb, ch, sqid=95):
    """Returns the PDB sequence cluster that contains chain *ch* in structure
    *pdb* for sequence identity level *sqid*.  PDB sequence cluster will be
    returned in as a list of tuples, e.g. ``[('1XXX', 'A'), ]``.  Note that
    PDB clusters individual chains, so the same PDB identifier may appear
    twice in the same cluster if the corresponding chain is present in the
    structure twice.

    Before this function is used, :func:`fetchPDBClusters` needs to be called.
    This function will load the PDB sequence clusters for *sqid* automatically
    using :func:`loadPDBClusters`."""

    assert isinstance(pdb, str) and len(pdb) == 4, \
        'pdb must be 4 char long string'
    assert isinstance(ch, str) and len(ch) == 1, \
        'ch must be a one char long string'
    try:
        sqid = int(sqid)
    except TypeError:
        raise TypeError('sqid must be an integer')
    if not (30 <= sqid <= 100):
        raise ValueError('sqid must be between 30 and 100')
    sqid = PDB_CLUSTERS_SQIDS[abs(PDB_CLUSTERS_SQIDS-sqid).argmin()]
    clusters = PDB_CLUSTERS[sqid]
    if clusters is None:
        loadPDBClusters(sqid)
        clusters = PDB_CLUSTERS[sqid]
    pdb_ch = (pdb.upper(), ch)

    for cluster in clusters:
        if pdb_ch in cluster:
            return cluster
    return 

def fetchPDBClusters(sqid=None):
    """Retrieve PDB sequence clusters.  PDB sequence clusters are results of
    the weekly clustering of protein chains in the PDB generated by blastclust.
    They are available at FTP site: https://cdn.rcsb.org/resources/sequence/clusters/

    This function will download about 10 Mb of data and save it after
    compressing in your home directory in :file:`.prody/pdbclusters`.
    Compressed files will be less than 4 Mb in size.  Cluster data can
    be loaded using :func:`loadPDBClusters` function and be accessed
    using :func:`listPDBCluster`."""

    if sqid is not None:
        if isListLike(sqid):
            for s in sqid:
                if s not in PDB_CLUSTERS:
                    raise ValueError('sqid must be one or more of ' + PDB_CLUSTERS_SQID_STR)
            keys = list(sqid)
        else:
            if sqid not in PDB_CLUSTERS:
                raise ValueError('sqid must be one or more of ' + PDB_CLUSTERS_SQID_STR)
            keys = [sqid]
    else:
        keys = list(PDB_CLUSTERS)

    PDB_CLUSTERS_PATH = os.path.join(getPackagePath(), 'pdbclusters')
    if not os.path.isdir(PDB_CLUSTERS_PATH):
        os.mkdir(PDB_CLUSTERS_PATH)
    LOGGER.progress('Downloading sequence clusters', len(keys),
                    '_prody_fetchPDBClusters')
    count = 0
    for i, x in enumerate(keys):
        filename = 'bc-{0}.out'.format(x)
        url = ('https://cdn.rcsb.org/resources/sequence/clusters/' + filename)
        try:
            inp = openURL(url)
        except IOError:
            LOGGER.warning('Clusters at {0}% sequence identity level could '
                           'not be downloaded.'.format(x))
            continue
        else:
            out = openFile(filename+'.gz', 'w', folder=PDB_CLUSTERS_PATH)
            out.write(inp.read())
            inp.close()
            out.close()
            count += 1
        LOGGER.update(i, label='_prody_fetchPDBClusters')
    LOGGER.finish()
    if len(keys) == count:
        LOGGER.info('All selected PDB clusters were downloaded successfully.')
    elif count == 0:
        LOGGER.warn('PDB clusters could not be downloaded.')



# -*- coding: utf-8 -*-
"""This module contains features for analyzing protein sequences.

Classes
=======

  * :class:`.MSA` - store MSA data indexed by label
  * :class:`.Sequence` - store sequence data


MSA IO
======

  * :class:`.MSAFile` - read/write MSA files in FASTA/SELEX/Stockholm formats
  * :func:`.parseMSA` - parse MSA files
  * :func:`.writeMSA` - parse MSA files

Editing
========

  * :func:`.mergeMSA` - merge MSA data for multi-domain proteins
  * :func:`.refineMSA` - refine MSA by removing gapped columns and/or sequences

Analysis
========

  * :func:`.calcMSAOccupancy` - calculate row (sequence) or column occupancy
  * :func:`.calcShannonEntropy` - calculate Shannon entropy
  * :func:`.buildMutinfoMatrix` - build mutual information matrix
  * :func:`.buildOMESMatrix` - build mutual observed minus expected squared
    covariance matrix
  * :func:`.buildSCAMatrix`- build statistical coupling analysis matrix
  * :func:`.buildSeqidMatrix`- build sequence identity matrix
  * :func:`.buildDirectInfoMatrix` - build direct information matrix
  * :func:`.uniqueSequences` - select unique sequences
  * :func:`.applyMutinfoCorr` - apply correction to mutual information matrix
  * :func:`.applyMutinfoNorm` - apply normalization to mutual information
    matrix
  * :func:`.calcMeff` - calculate sequence weights
  * :func:`.calcRankorder` - rank order scores


Plotting
========

  * :func:`.showShannonEntropy` - plot Shannon entropy
  * :func:`.showMSAOccupancy` - plot row (sequence) or column occupancy
  * :func:`.showMutinfoMatrix` - show mutual information matrix
"""

__all__ = []

from . import msa
from .msa import *
__all__.extend(msa.__all__)

from . import msafile
from .msafile import *
__all__.extend(msafile.__all__)

from . import analysis
from .analysis import *
__all__.extend(analysis.__all__)

from . import plotting
from .plotting import *
__all__.extend(plotting.__all__)

from . import sequence
from .sequence import *
__all__.extend(sequence.__all__)




# -*- coding: utf-8 -*-
""" This module defines a class for identifying contacts."""

import numpy as np

from prody import LOGGER
from prody.atomic import AtomPointer, AtomMap
from prody.utilities import importLA, checkWeights

from .measure import calcCenter

linalg = importLA()

__all__ = ['Transformation', 'applyTransformation', 'alignCoordsets',
           'calcRMSD', 'calcTransformation', 'superpose',
           'moveAtoms', 'wrapAtoms',
           'printRMSD']


class Transformation(object):

    """A class for storing a transformation matrix."""

    __slots__ = ['_matrix']

    def __init__(self, *args):
        """Either 4x4 transformation *matrix*, or *rotation* matrix and
        *translation* vector must be provided at instantiation."""

        nargs = len(args)
        if nargs == 1:
            self.setMatrix(args[0])
        elif nargs == 2:
            self._matrix = np.eye(4)
            self.setRotation(args[0])
            self.setTranslation(args[1])
        else:
            raise ValueError('one or two array must be provided as arguments')

    def getRotation(self):
        """Returns rotation matrix."""

        if self._matrix is not None:
            return self._matrix[:3, :3]

    def setRotation(self, rotation):
        """Set rotation matrix."""

        if not isinstance(rotation, np.ndarray):
            raise TypeError('rotation must be an ndarray')
        elif rotation.shape != (3, 3):
            raise ValueError('rotation must be a 3x3 array')
        self._matrix[:3, :3] = rotation

    def getTranslation(self):
        """Returns translation vector."""

        if self._matrix is not None:
            return self._matrix[:3, 3]

    def setTranslation(self, translation):
        """Set translation vector."""

        if not isinstance(translation, np.ndarray):
            raise TypeError('translation must be an ndarray')
        elif translation.shape != (3,):
            raise ValueError('translation must be an ndarray of length 3')
        self._matrix[:3, 3] = translation

    def getMatrix(self):
        """Returns a copy of the 4x4 transformation matrix whose top left is
        rotation matrix and last column is translation vector."""

        if self._matrix is not None:
            return self._matrix.copy()

    def setMatrix(self, matrix):

        if not isinstance(matrix, np.ndarray):
            raise TypeError('matrix must be an ndarray')
        elif matrix.shape != (4, 4):
            raise ValueError('matrix must have shape (4,4)')
        self._matrix = matrix

    def apply(self, atoms):
        """Apply transformation to *atoms*, see :func:`applyTransformation`
        for details."""

        return applyTransformation(self, atoms)


def calcTransformation(mobile, target, weights=None):
    """Returns a :class:`Transformation` instance which, when applied to the
    atoms in *mobile*, minimizes the weighted RMSD between *mobile* and
    *target*.  *mobile* and *target* may be NumPy coordinate arrays, or
    :class:`.Atomic` instances, e.g. :class:`.AtomGroup`, :class:`.Chain`,
    or :class:`.Selection`."""

    if not isinstance(mobile, np.ndarray):
        try:
            mob = mobile._getCoords()
        except AttributeError:
            raise TypeError('mobile must be a numpy array or an object '
                            'with getCoords method')
    else:
        mob = mobile
    if not isinstance(target, np.ndarray):
        try:
            tar = target._getCoords()
        except AttributeError:
            raise TypeError('target must be a numpy array or an object '
                            'with getCoords method')
    else:
        tar = target

    if mob.shape != tar.shape:
        raise ValueError('reference and target coordinate arrays '
                         'must have same number of atoms')

    if mob.shape[1] != 3:
        raise ValueError('reference and target must be coordinate arrays')
    
    if weights is None:
        if isinstance(mobile, AtomMap):
            LOGGER.warn('mobile is an AtomMap instance, consider assign weights=mobile.getFlags("mapped") '
                        'if there are dummy atoms in mobile')

        if isinstance(target, AtomMap):
            LOGGER.warn('target is an AtomMap instance, consider assign weights=target.getFlags("mapped") '
                        'if there are dummy atoms in target')

    if weights is not None:
        weights = checkWeights(weights, mob.shape[0])

    return Transformation(*getTransformation(mob, tar, weights))


def getTransformation(mob, tar, weights=None):

    if weights is None:
        mob_com = mob.mean(0)
        tar_com = tar.mean(0)
        mob = mob - mob_com
        tar = tar - tar_com
        matrix = np.dot(mob.T, tar)
    else:
        weights_sum = weights.sum()
        weights_dot = np.dot(weights.T, weights)
        mob_com = (mob * weights).sum(axis=0) / weights_sum
        tar_com = (tar * weights).sum(axis=0) / weights_sum
        mob = mob - mob_com
        tar = tar - tar_com
        matrix = np.dot((mob * weights).T, (tar * weights)) / weights_dot

    U, _, Vh = linalg.svd(matrix)
    d = np.sign(linalg.det(np.dot(U, Vh)))
    Id = np.array([[1, 0, 0],
                   [0, 1, 0],
                   [0, 0, d]])
    rotation = np.dot(Vh.T, np.dot(Id, U.T))

    return rotation, tar_com - np.dot(mob_com, rotation.T)


def applyTransformation(transformation, atoms):
    """Returns *atoms* after applying *transformation*.  If *atoms*
    is a :class:`.Atomic` instance, it will be returned after
    *transformation* is applied to its active coordinate set.  If
    *atoms* is an :class:`.AtomPointer` instance, *transformation*
    will be applied to the corresponding coordinate set in the
    associated :class:`.AtomGroup`."""

    coords = None
    ag = None
    if isinstance(atoms, np.ndarray):
        atoms = np.atleast_2d(atoms)
        if atoms.shape[1] != 3:
            raise ValueError('atoms must be a 3-d coordinate array')
        coords = atoms
        atoms = None
    else:
        if isinstance(atoms, AtomPointer):
            ag = atoms.getAtomGroup()
            acsi = ag.getACSIndex()
            ag.setACSIndex(atoms.getACSIndex())
            coords = ag._getCoords()
        else:
            try:
                coords = atoms._getCoords()
            except AttributeError:
                raise TypeError('atoms must be a Atomic instance')

    if atoms is None:
        return _applyTransformation(transformation, coords)
    else:
        if ag is None:
            atoms.setCoords(_applyTransformation(transformation, coords))
        else:
            ag.setCoords(_applyTransformation(transformation, coords))
            ag.setACSIndex(acsi)
        return atoms


def _applyTransformation(t, coords):

    return np.dot(coords, t.getRotation().T) + t.getTranslation()


def superpose(mobile, target, weights=None):
    """Returns *mobile*, after its RMSD minimizing superposition onto *target*,
    and the transformation that minimizes the RMSD."""

    t = calcTransformation(mobile, target, weights)
    result = applyTransformation(t, mobile)
    return (result, t)


def moveAtoms(atoms, **kwargs):
    """Move *atoms* *to* a new location or *by* an offset.  This method will
    change the active coordinate set of the *atoms*.  Note that only one of
    *to* or *by* keyword arguments is expected.

    Move protein so that its centroid is at the origin, ``[0., 0., 0.]``:

    .. ipython:: python

       from prody import *
       from numpy import ones, zeros
       protein = parsePDB('1ubi')
       calcCenter(protein).round(3)
       moveAtoms(protein, to=zeros(3))
       calcCenter(protein).round(3)

    Move protein so that its mass center is at the origin:

    .. ipython:: python

       protein.setMasses(ones(len(protein)))
       protein.carbon.setMasses(12)
       protein.nitrogen.setMasses(14)
       protein.oxygen.setMasses(16)
       moveAtoms(protein, to=zeros(3), weights=protein.getMasses())
       calcCenter(protein, weights=protein.getMasses()).round(3)

    Move protein so that centroid of Cα atoms is at the origin:

    .. ipython:: python

       moveAtoms(protein.ca, to=zeros(3), ag=True)
       calcCenter(protein).round(3)
       calcCenter(protein.ca).round(3)

    Move protein by 10 A along each direction:

    .. ipython:: python

       moveAtoms(protein, by=ones(3) * 10)
       calcCenter(protein).round(3)
       calcCenter(protein.ca).round(3)


    :arg by: an offset array with shape ``([1,] 3)`` or ``(n_atoms, 3)`` or
        a transformation matrix with shape ``(4, 4)``
    :type by: :class:`numpy.ndarray`

    :arg to: a point array with shape ``([1,] 3)``
    :type to: :class:`numpy.ndarray`

    :arg ag: when *atoms* is a :class:`.AtomSubset`, apply translation vector
        (*to*) or transformation matrix to the :class:`.AtomGroup`,
        default is **False**
    :type ag: bool

    :arg weights: array of atomic weights with shape ``(n_atoms[, 1])``
    :type weights: :class:`numpy.ndarray`

    When *to* argument is passed, :func:`.calcCenter` function is used to
    calculate centroid or mass center."""

    point = kwargs.pop('to', None)
    if point is None:
        offset = kwargs.pop('by', None)
        if offset is None:
            raise TypeError('moveAtoms() expects one of {0} or {1} '
                            'arguments'.format(repr('to'), repr('by')))

        try:
            shape = offset.shape
        except AttributeError:
            raise TypeError('by must be a numpy array')

        if shape == (4, 4):
            if kwargs.pop('ag', False):
                try:
                    atoms = atoms.getAtomGroup()
                except AttributeError:
                    # itself must be an AtomGroup
                    pass
            try:
                coords = atoms._getCoords()
            except AttributeError:
                try:
                    coords = atoms.getCoords()
                except AttributeError:
                    raise TypeError('atoms must be an Atomic instance')

            coords = np.dot(coords, offset[:3, :3])
            coords += offset[3, :3]

            atoms.setCoords(coords)
        else:

            try:
                coords = atoms._getCoords()
            except AttributeError:
                try:
                    coords = atoms.getCoords()
                except AttributeError:
                    raise TypeError('atoms must be an Atomic instance')

            try:
                natoms = atoms.numAtoms()
            except AttributeError:
                raise TypeError('atoms must be an Atomic instance')
            if shape == (3,) or shape == (1, 3) or shape == (natoms, 3):
                atoms.setCoords(coords + offset)
            else:
                raise ValueError('by.shape is not valid')

    else:
        try:
            shape = point.shape
        except AttributeError:
            raise TypeError('to must be a numpy array')
        if shape != (3,) and shape != (1, 3):
            raise ValueError('to.shape must be ([1,] 3)')

        center = calcCenter(atoms, weights=kwargs.pop('weights', None))
        offset = point - center
        if kwargs.pop('ag', False):
            try:
                atoms = atoms.getAtomGroup()
            except AttributeError:
                # itself must be an AtomGroup
                pass

        try:
            coords = atoms._getCoords()
        except AttributeError:
            try:
                coords = atoms.getCoords()
            except AttributeError:
                raise TypeError('atoms must be an Atomic instance')
        atoms.setCoords(coords + offset)


def calcRMSD(reference, target=None, weights=None):
    """Returns root-mean-square deviation (RMSD) between reference and target
    coordinates.

    .. ipython:: python

       ens = loadEnsemble('p38_X-ray.ens.npz')
       ens.getRMSDs()
       calcRMSD(ens)
       calcRMSD(ens.getCoords(), ens.getCoordsets(), ens.getWeights())"""

    if isinstance(reference, np.ndarray):
        ref = reference
    else:
        try:
            ref = reference._getCoords()
        except AttributeError:
            raise TypeError('reference must be a numpy array or an object '
                            'with getCoords method')
        if target is None:
            try:
                target = reference._getCoordsets()
            except AttributeError:
                pass
        if weights is None:
            try:
                weights = reference._getWeights()
            except AttributeError:
                pass
    if ref.ndim != 2 or ref.shape[1] != 3:
        raise ValueError('reference must have shape (n_atoms, 3)')

    if isinstance(target, np.ndarray):
        tar = target
    else:
        try:
            tar = target._getCoords()
        except AttributeError:
            raise TypeError('target must be a numpy array or an object '
                            'with getCoords method')
    if tar.ndim not in (2, 3) or tar.shape[-1] != 3:
        raise ValueError('target must have shape ([n_confs,] n_atoms, 3)')

    if ref.shape != tar.shape[-2:]:
        raise ValueError('reference and target arrays must have the same '
                         'number of atoms')

    if weights is not None:
        n_atoms = len(ref)
        n_csets = 1 if tar.ndim == 2 else len(tar)
        weights = checkWeights(weights, n_atoms, n_csets)
    return getRMSD(ref, tar, weights)


def getRMSD(ref, tar, weights=None):
    if weights is None:
        divByN = 1.0 / ref.shape[0]
        if tar.ndim == 2:
            return np.sqrt(((ref-tar) ** 2).sum() * divByN)
        else:
            rmsd = np.zeros(len(tar))
            for i, t in enumerate(tar):
                rmsd[i] = ((ref-t) ** 2).sum()
            return np.sqrt(rmsd * divByN)
    else:
        if tar.ndim == 2:
            return np.sqrt((((ref-tar) ** 2) * weights).sum() *
                           (1. / weights.sum()))
        else:
            rmsd = np.zeros(len(tar))
            if weights.ndim == 2:
                for i, t in enumerate(tar):
                    rmsd[i] = (((ref-t) ** 2) * weights).sum()
                return np.sqrt(rmsd * (1. / weights.sum()))
            else:
                for i, t in enumerate(tar):
                    rmsd[i] = (((ref-t) ** 2) * weights[i]).sum()
                return np.sqrt(rmsd / weights.sum(1).flatten())


def printRMSD(reference, target=None, weights=None, log=True, msg=None):
    """Print RMSD to the screen.  If *target* has multiple coordinate sets,
    minimum, maximum and mean RMSD values are printed.  If *log* is **True**
    (default), RMSD is written to the standard error using package logger,
    otherwise standard output is used.  When *msg* string is given, it is
    printed before the RMSD value.  See also :func:`calcRMSD` function. """

    if log:
        write = LOGGER.info
    else:
        import sys
        write = lambda line: sys.stdout.write(line + '\n')
    msg = msg or ''
    if msg and msg[-1] != ' ':
        msg += ' '
    rmsd = calcRMSD(reference, target, weights)
    if isinstance(rmsd, np.ndarray) and len(rmsd) > 1:
        write(msg + 'RMSD: min={0:.2f}, max={1:.2f}, mean={2:.2f}'
                    .format(rmsd.min(), rmsd.max(), rmsd.mean()))
    else:
        if isinstance(rmsd, np.ndarray):
            rmsd = rmsd[0]
        write(msg + 'RMSD: {0:.2f}'.format(rmsd))


def alignCoordsets(atoms, weights=None):
    """Returns *atoms* after superposing coordinate sets onto its active
    coordinate set.  Transformations will be calculated for *atoms* and
    applied to its :class:`.AtomGroup`, when applicable.  Optionally,
    atomic *weights* can be passed for weighted superposition."""

    try:
        acsi, n_csets = atoms.getACSIndex(), atoms.numCoordsets()
    except AttributeError:
        raise TypeError('atoms must have type Atomic, not {0}'
                        .format(type(atoms)))
        if n_csets < 2:
            LOGGER.warning('{0} contains fewer than two coordinate sets, '
                           'alignment was not performed.'.format(str(atoms)))
            return

    try:
        ag = atoms.getAtomGroup()
    except AttributeError:
        ag = atoms
    agacsi = ag.getACSIndex()

    tar = atoms._getCoords()
    for i in range(n_csets):
        if i == acsi:
            continue
        atoms.setACSIndex(i)
        ag.setACSIndex(i)
        calcTransformation(atoms, tar, weights).apply(ag)
    atoms.setACSIndex(acsi)
    ag.setACSIndex(agacsi)
    return atoms


def wrapAtoms(frame, unitcell=None, center=np.array([0., 0., 0.])):
    """Wrap atoms into an image of the system simulated under periodic boundary
    conditions. When *frame* is a :class:`.Frame`, unitcell information will be
    retrieved automatically.

    .. note::
       This function will wrap all atoms into the specified periodic image, so
       covalent bonds will be broken.

    :arg frame: a frame instance or a coordinate set
    :type frame: :class:`.Frame`, :class:`.AtomGroup`, :class:`numpy.ndarray`

    :arg unitcell: orthorhombic unitcell array with shape (3,)
    :type unitcell: :class:`numpy.ndarray`

    :arg center: coordinates of the center of the wrapping cell, default is
        the origin of the Cartesian coordinate system
    :type center: :class:`numpy.ndarray`"""

    try:
        coords = frame._getCoords()
    except AttributeError:
        coords = frame
    else:
        try:
            frame.getAtomGroup()
        except AttributeError:
            pass
        else:
            raise TypeError('frame must be a Frame, AtomGroup, or numpy array,'
                            ' not a ' + str(type(frame)))

    if unitcell is None:
        try:
            unitcell = frame.getUnitcell()[:3]
        except AttributeError:
            raise TypeError('unitcell information must be provided')

    half = unitcell / 2
    ucmin = center - half
    ucmax = center + half
    for axis in range(3):
        xyz = coords[:, axis]
        which = (xyz < ucmin[axis]).nonzero()[0]
        while len(which):
            coords[which, axis] += unitcell[axis]
            which = which[coords[which, axis] < ucmin[axis]]
        which = (xyz > ucmax[axis]).nonzero()[0]
        while len(which):
            coords[which, axis] -= unitcell[axis]
            which = which[coords[which, axis] > ucmax[axis]]
    return frame



# -*- coding: utf-8 -*-
'''
Copyright (c) 2020-2022 Burak Kaynak, Pemra Doruker.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

__author__ = 'Anupam Banerjee'
__credits__ = ['James Krieger']
__email__ = ['anupam.banerjee@stonybrook.edu', 'jamesmkrieger@gmail.com']

from numbers import Number
import os

from prody import LOGGER
from prody.atomic.atomic import Atomic
from prody.ensemble.ensemble import Ensemble
from prody.proteins.pdbfile import parsePDB, writePDB

from prody.dynamics.anm import ANM
from prody.dynamics.clustenm import ClustENM
from prody.dynamics.editing import extendModel
from prody.dynamics.modeset import ModeSet
from prody.dynamics.nma import NMA
from prody.dynamics.pca import PCA
from prody.dynamics.sampling import traverseMode


__all__ = ['runANMD']

def runANMD(atoms, num_modes=2, max_rmsd=2., num_steps=5, tolerance=10.0,
            **kwargs):
    """Runs the ANMD hybrid simulation method ([CM22]_), which generates conformations
    along single modes using :func:`.traverseModes` and minimises them. 
    
    The first non-zero mode is scaled to *max_rmsd*
    and the remaining modes are scaled accordingly using their eigenvalues.

    kwargs of traverseMode can be provided: pos, neg and reverse

    :arg atoms: an object with atom and coordinate data.
        This should be a complete atomic model. It is ok to be missing 
        some side chain atoms and hydrogens, but not fragments, such as loops.
    :type atoms: :class:`.Atomic`

    :arg num_modes: number of modes to calculate
        Default is 2
    :type num_modes: int

    :arg max_rmsd: maximum rmsd for non-zero mode 1
        Default is 2.
    :type max_rmsd: float

    :arg num_steps: number of conformers in each direction for each mode
        Default is 5
    :type num_steps: int

    :arg tolerance: tolerance for energy minimisation in OpenMM 
        in kilojoule/mole/nanometer. Default is 10 as in OpenMM
    :type tolerance: float

    :arg skip_modes: number of modes to skip
        Default is 0
    :type skip_modes: int

    :arg anm: your own NMA modes or ModeSet to use instead
        Default is None
    :type anm: :class:`.NMA`, :class:`.ANM`, :class:`.ModeSet`

    .. [CM22] Mary Hongying Cheng, James M Krieger, Anupam Banerjee, Yufei Xiang, 
            Burak Kaynak, Yi Shi, Moshe Arditi, Ivet Bahar. 
            Impact of new variants on SARS-CoV-2 infectivity and neutralization: 
            A molecular assessment of the alterations in the spike-host protein 
            interactions. *iScience* **2022** 25(3):103939.
    """
    try:
        from simtk.openmm.app import PDBFile, ForceField, \
            Simulation, HBonds, NoCutoff
        from simtk.openmm import LangevinIntegrator
        from simtk.unit import nanometer, kelvin, picosecond, picoseconds, \
            angstrom, kilojoule, mole
    except ImportError:
        raise ImportError('Please install PDBFixer and OpenMM to use ANMD')

    if not isinstance(atoms, Atomic):
        raise TypeError('atoms should be an Atomic object')

    if not isinstance(num_modes, int):
        raise TypeError('num_modes should be an integer')

    if not isinstance(num_steps, int):
        raise TypeError('num_steps should be an integer')

    if not isinstance(max_rmsd, Number):
        raise TypeError('max_rmsd should be a float')

    if not isinstance(tolerance, Number):
        raise TypeError('tolerance should be a float')
    tolerance = tolerance * kilojoule/mole/nanometer

    pos = kwargs.get('pos', True)
    if not isinstance(pos, bool):
        raise TypeError('pos should be a bool')
    
    neg = kwargs.get('neg', True)
    if not isinstance(neg, bool):
        raise TypeError('neg should be a bool')
    
    reverse = kwargs.get('reverse', False)
    if not isinstance(reverse, bool):
        raise TypeError('reverse should be a bool')

    skip_modes = kwargs.get('skip_modes', 0)
    if not isinstance(skip_modes, int):
        raise TypeError('skip_modes should be an integer')

    anm = kwargs.get('anm', None)
    if not isinstance(anm, (type(None), NMA, ModeSet)):
        raise TypeError('anm should be an NMA or ModeSet object')

    pdb_name=atoms.getTitle().replace(' ', '_')

    fix_name = pdb_name + '_fixed.pdb'
    if os.path.exists(fix_name):
        LOGGER.info('\nFixed structure found')
    else:
        clustenm=ClustENM()
        clustenm.setAtoms(atoms)
        clustenm.writePDBFixed()
    pdb_fix = PDBFile(fix_name)

    fixmin_name=pdb_name + '_fixedmin.pdb'
    if os.path.exists(fixmin_name):
        LOGGER.info('\nMinimised fixed structure found')
    else:
        LOGGER.info('\nMinimising fixed structure ...')
        LOGGER.timeit('_anmd_min')

        forcefield = ForceField("amber99sbildn.xml", "amber99_obc.xml")
        system = forcefield.createSystem(pdb_fix.topology, 
                                         nonbondedMethod=NoCutoff,
                                         nonbondedCutoff=1*nanometer, 
                                         constraints=HBonds)
        integrator = LangevinIntegrator(300*kelvin, 1/picosecond, 
                                        0.003*picoseconds)
        simulation = Simulation(pdb_fix.topology, system, integrator)
        simulation.context.setPositions(pdb_fix.positions)
        simulation.minimizeEnergy(tolerance=tolerance)
        positions = simulation.context.getState(getPositions=True).getPositions()
        PDBFile.writeFile(simulation.topology, positions, open(fixmin_name, 'w'))
        LOGGER.report('The fixed structure was minimised in %.2fs.\n',
                        label='_anmd_min')

    pdb_fixed=parsePDB(fixmin_name, compressed=False)

    if skip_modes >= num_modes:
        LOGGER.warn('skip_modes >= num_modes so no modes are used and minimised fixed structure is returned')
        return pdb_fixed

    calphas=pdb_fixed.select('calpha')
    if anm is None:
        anm=ANM()
        anm.buildHessian(calphas)
        anm.calcModes(n_modes=num_modes)
    elif isinstance(anm, PCA):
        eigvecs = anm.getEigvecs()
        eigvals = 1/anm.getEigvals()
        anm = ANM()
        anm.setEigens(eigvecs, eigvals)

    anm_ex, atoms_all = extendModel(anm, calphas, pdb_fixed)
    anm_ex._indices = anm.getIndices()
    eval_0=anm[0].getEigval()

    ensembles = []
    for i in range(skip_modes, num_modes):
        modeNum = anm_ex.getIndices()[i]

        eval_i=anm[i].getEigval()
        sc_rmsd=((1/eval_i)**0.5/(1/eval_0)**0.5)*max_rmsd
        traj_aa=traverseMode(anm_ex[i], atoms_all, n_steps=num_steps, rmsd=sc_rmsd,
                             **kwargs)
        traj_aa.setAtoms(atoms_all)

        num_confs = traj_aa.numConfs()
        LOGGER.info('\nMinimising {0} conformers for mode {1} ...'.format(num_confs, modeNum))

        target_ensemble = Ensemble('mode {0} ensemble'.format(modeNum))
        target_ensemble.setAtoms(atoms_all)
        target_ensemble.setCoords(atoms_all)
        
        for j, conf in enumerate(traj_aa):
            jp1 = j+1
            writePDB('temp1.pdb', conf)
            pdb = PDBFile('temp1.pdb')
            os.remove("temp1.pdb")

            LOGGER.info('\nMinimising structure {0} along mode {1} ...'.format(jp1, modeNum))
            LOGGER.timeit('_anmd_min')
            forcefield = ForceField("amber99sbildn.xml", "amber99_obc.xml")
            system = forcefield.createSystem(pdb.topology, 
                                             nonbondedMethod=NoCutoff,
                                             nonbondedCutoff=1*nanometer, 
                                             constraints=HBonds)
            integrator = LangevinIntegrator(300*kelvin, 1/picosecond, \
                                            0.003*picoseconds)
            simulation = Simulation(pdb.topology, system, integrator)
            simulation.context.setPositions(pdb.positions)
            simulation.minimizeEnergy(tolerance=tolerance)
            positions = simulation.context.getState(getPositions=True).getPositions(
                asNumpy=True).value_in_unit(angstrom)[:pdb.topology.getNumAtoms()]
            
            target_ensemble.addCoordset(positions)

            LOGGER.report('The structure was minimised in %.2fs.',
                            label='_anmd_min')
            
        ensembles.append(target_ensemble)

    os.remove(fix_name)
    os.remove(fixmin_name)

    return ensembles


if __name__=='__main__':
    import sys
    from prody.tests.datafiles import pathDatafile

    pdb_filename = sys.argv[1] if len(sys.argv) > 1 else pathDatafile('1ubi')
    num_modes = int(sys.argv[2]) if len(sys.argv) > 2 else 2
    max_rmsd = float(sys.argv[3]) if len(sys.argv) > 3 else 2.
    tol = float(sys.argv[4]) if len(sys.argv) > 4 else 10.

    num_steps = int(sys.argv[5]) if len(sys.argv) > 5 else 2
    skip_modes = int(sys.argv[6]) if len(sys.argv) > 6 else 0

    pos = bool(sys.argv[7]) if len(sys.argv) > 7 else True
    neg = bool(sys.argv[8]) if len(sys.argv) > 8 else True
    reverse = bool(sys.argv[9]) if len(sys.argv) > 9 else False

    anm_filename = sys.argv[10] if len(sys.argv) > 10 else None
    anm = None
    if anm_filename is not None:
        if anm.endswith('nmd'):
            anm, _ = parseNMD(anm_filename)
        elif anm.endswith('npz'):
            anm = loadModel(anm_filename)
        else:
            raise ValueError('anm should be an nmd or npz file or None')

    pdb_name_ext = pdb_filename
    if pdb_name_ext.endswith('.pdb'):
        pdb_filename = pdb_name_ext[:-4]
    else:
        pdb_name_ext += '.pdb'

    pdb = parsePDB(pdb_name_ext, compressed=False)

    x = runANMD(pdb, num_modes=num_modes, max_rmsd=max_rmsd,
                num_steps=num_steps, skip_modes=skip_modes, tolerance=tol,
                pos=pos, neg=neg, reverse=reverse,
                anm=anm)

    pdb_basename = os.path.basename(pdb_filename)
    for ens in x:
        filename = pdb_basename + '_' + ens.getTitle().replace(' ', '_')
        LOGGER.info('writing PDB file {0}'.format(filename))
        writePDB(filename, ens)



"""This module defines structure and dynamics analysis applications."""

import importlib
import sys

try:
    import argparse
except ImportError:
    from .. import argparse

from prody.apps.apptools import *
from prody.utilities.misctools import impLoadModule

if sys.version_info[0] == 2:
    import imp
    path_prody = imp.find_module('prody')[1]
else:
    path_prody = importlib.util.find_spec("prody").submodule_search_locations[0]

try:
    import imp
    path_apps = imp.find_module('apps', [path_prody])[1]
    path_apps = imp.find_module('prody_apps', [path_apps])[1]
except ModuleNotFoundError:
    path_apps = importlib.util.find_spec("prody.apps").submodule_search_locations[0]
    path_apps += '/prody_apps/'

PRODY_APPS = ['anm', 'gnm', 'pca', 'eda', 'align', 'blast', 'biomol',
                  'catdcd', 'contacts', 'fetch', 'select', 'energy', 
                  'clustenm', 'rtb']

__all__ = ['prody_main']

prody_parser = argparse.ArgumentParser(
    description="ProDy: A Python Package for Protein Dynamics Analysis",
    epilog="See 'prody <command> -h' for more information on a specific "
           "command."
    )

prody_parser.add_argument('-c', '--cite',
    help="print citation info and exit",
    action=ProDyCitation, nargs=0)

prody_parser.add_argument('-v', '--version',
    help="print ProDy version and exit",
    action=ProDyVersion, nargs=0)

prody_commands = prody_parser.add_subparsers(
    title='subcommands')


for cmd in PRODY_APPS:
    cmd = 'prody_' + cmd
    mod = impLoadModule('prody.apps.prody_apps.', cmd, path_apps)
    mod.addCommand(prody_commands)

def prody_main():

    if len(sys.argv) == 1:
        prody_parser.print_help()
    else:
        namespace = prody_parser.parse_args()
        namespace.func(namespace)


if __name__ == '__main__':
    prody_main()



# -*- coding: utf-8 -*-
"""This module defines customized gamma functions for elastic network model
analysis."""

import numpy as np

from prody.atomic import Atomic

__all__ = ['Gamma', 'GammaStructureBased', 'GammaVariableCutoff', 'GammaED', 'GammaGOdMD']


class Gamma(object):

    """Base class for facilitating use of atom type, residue type, or residue
    property dependent force constants (γ).

    Derived classes:

    * :class:`.GammaStructureBased`
    * :class:`.GammaVariableCutoff`"""

    def __init__(self):
        pass

    def gamma(self, dist2, i, j):
        """Returns force constant.

        For efficiency purposes square of the distance between interacting
        atom/residue (node) pairs is passed to this function. In addition,
        node indices are passed."""

        pass


class GammaStructureBased(Gamma):

    """Facilitate setting the spring constant based on the secondary structure
    and connectivity of the residues.

    A systematic study [LT10]_ of a large set of NMR-structures analyzed
    using a method based on entropy maximization showed that taking into
    consideration properties such as sequential separation between
    contacting residues and the secondary structure types of the interacting
    residues provides refinement in the ENM description of proteins.

    This class determines pairs of connected residues or pairs of proximal
    residues in a helix or a sheet, and assigns them a larger user defined
    spring constant value.

     DSSP single letter abbreviations are recognized:
       * **H**: α-helix
       * **G**: 3-10-helix
       * **I**: π-helix
       * **E**: extended part of a sheet

    *helix*:
        Applies to residue (or Cα atom) pairs that are in the same helical
        segment, at most 7 Å apart, and separated by at most
        3 (3-10-helix), 4 (α-helix), or 5 (π-helix) residues.

    *sheet*:
        Applies to Cα atom pairs that are in different β-strands and at most
        6 Å apart.

    *connected*:
        Applies to Cα atoms that are at most 4 Å apart.

    Note that this class does not take into account insertion codes.

    .. [LT10] Lezon TR, Bahar I. Using entropy maximization to understand the
       determinants of structural dynamics beyond native contact topology.
       *PLoS Comput Biol* **2010** 6(6):e1000816.

    **Example**:

    Let's parse coordinates and header data from a PDB file, and then
    assign secondary structure to the atoms.

    .. ipython:: python

       from prody import *
       ubi, header = parsePDB('1aar', chain='A', subset='calpha', header=True)
       assignSecstr(header, ubi)

    In the above we parsed only the atoms needed for this calculation, i.e.
    Cα atoms from chain A.

    We build the Hessian matrix using structure based force constants as
    follows;

    .. ipython:: python

       gamma = GammaStructureBased(ubi)
       anm = ANM('')
       anm.buildHessian(ubi, gamma=gamma)

    We can obtain the force constants assigned to residue pairs from the
    Kirchhoff matrix as follows:

    .. ipython:: python

       k = anm.getKirchhoff()
       k[0,1] # a pair of connected residues
       k[0,16] # a pair of residues from a sheet"""

    def __init__(self, atoms, gamma=1.0, helix=6.0, sheet=6.0, connected=10.0):
        """Setup the parameters.

        :arg atoms: A set of atoms with chain identifiers, residue numbers,
            and secondary structure assignments are set.
        :type atoms: :class:`.Atomic`

        :arg gamma: Force constant in arbitrary units. Default is 1.0.
        :type gamma: float

        :arg helix: Force constant factor for residues hydrogen bonded in
            α-helices, 3,10-helices, and π-helices. Default is 6.0, i.e.
            ``6.0`*gamma``.
        :type helix: float

        :arg sheet: Force constant factor for residue pairs forming a hydrogen
            bond in a β-sheet. Default is 6.0, i.e. ``6.0`*gamma``.
        :type sheet: float

        :arg connected: Force constant factor for residue pairs that are
            connected. Default is 10.0, i.e. ``10.0`*gamma``.
        :type connected: float"""

        if not isinstance(atoms, Atomic):
            raise TypeError('atoms must be an Atomic instance')
        n_atoms = atoms.numAtoms()
        if n_atoms < 3:
            raise ValueError('number of atoms must be larger than 2')
        sstr = atoms.getSecstrs()
        assert sstr is not None, 'secondary structure assignments must be set'
        chid = atoms.getChids()
        assert chid is not None, 'chain identifiers must be set'
        rnum = atoms.getResindices()
        assert rnum is not None, 'residue numbers must be set'
        gamma = float(gamma)
        assert gamma > 0, 'gamma must be greater than 0'
        helix = float(helix)
        assert helix > 0, 'helix must be greater than 0'
        sheet = float(sheet)
        assert sheet > 0, 'sheet must be greater than 0'
        connected = float(connected)
        assert connected > 0, 'connected must be greater than 0'

        ssid = np.zeros(n_atoms)
        for i in range(1, n_atoms):
            if (sstr[i-1] == sstr[i] and chid[i-1] == chid[i] and
               rnum[i]-rnum[i-1] == 1):
                ssid[i] = ssid[i-1]
            else:
                ssid[i] = ssid[i-1] + 1
        self._sstr = sstr
        self._chid = chid
        self._rnum = rnum
        self._ssid = ssid
        self._gamma = gamma
        self._helix = gamma * helix
        self._sheet = gamma * sheet
        self._connected = gamma * connected

    def getSecstrs(self):
        """Returns a copy of secondary structure assignments."""

        return self._sstr.copy()

    def getChids(self):
        """Returns a copy of chain identifiers."""

        return self._chid.socopypy()

    def getResnums(self):
        """Returns a copy of residue numbers."""

        return self._rnum.copy()

    def gamma(self, dist2, i, j):
        """Returns force constant."""

        if dist2 <= 16:
            return self._connected
        sstr = self._sstr
        ssid = self._ssid
        rnum = self._rnum
        # if residues are in the same secondary structure element
        if ssid[i] == ssid[j]:
            i_j = abs(rnum[j] - rnum[i])
            if ((i_j <= 4 and sstr[i] == 'H') or
                    (i_j <= 3 and sstr[i] == 'G') or
                    (i_j <= 5 and sstr[i] == 'I')) and dist2 <= 49:
                return self._helix
        elif sstr[i] == sstr[j] == 'E' and dist2 <= 36:
            return self._sheet

        return self._gamma


class GammaVariableCutoff(Gamma):

    """Facilitate setting the cutoff distance based on user defined
    atom/residue (node) radii.

    Half of the cutoff distance can be thought of as the radius of a node.
    This class enables setting different radii for different node types.

    **Example**:

    Let's think of a protein-DNA complex for which we want to use different
    radius for different residue types. Let's say, for protein Cα atoms we
    want to set the radius to 7.5 Å, and for nucleic acid phosphate atoms to
    10 Å. We use the HhaI-DNA complex structure :file:`1mht`.

    .. ipython:: python

       hhai = parsePDB('1mht')
       ca_p = hhai.select('(protein and name CA) or (nucleic and name P)')
       ca_p.getNames()

    We set the radii of atoms:

    .. ipython:: python

       varcutoff = GammaVariableCutoff(ca_p.getNames(), gamma=1,
           default_radius=7.5, debug=False, P=10)
       varcutoff.getRadii()

    The above shows that for phosphate atoms radii is set to 10 Å, because
    we passed the ``P=10`` argument.  As for Cα atoms, the default 7.5 Å
    is set as the radius (``default_radius=7.5``).  You can also try this with
    ``debug=True`` argument to print debugging information on the screen.

    We build :class:`.ANM` Hessian matrix as follows:

    .. ipython:: python

       anm = ANM('HhaI-DNA')
       anm.buildHessian(ca_p, gamma=varcutoff, cutoff=20)

    Note that we passed ``cutoff=20.0`` to the :meth:`.ANM.buildHessian`
    method.  This is equal to the largest possible cutoff distance (between
    two phosphate atoms) for this system, and ensures that all of the
    potential interactions are evaluated.

    For pairs of atoms for which the actual distance is larger than the
    effective cutoff, the :meth:`.GammaVariableCutoff.gamma` method returns
    ``0``.  This annuls the interaction between those atom pairs."""

    def __init__(self, identifiers, gamma=1., default_radius=7.5, **kwargs):
        """Set the radii of atoms.

        :arg identifiers: List of atom names or types, or residue names.
        :type identifiers: list or :class:`numpy.ndarray`

        :arg gamma: Uniform force constant value. Default is 1.0.
        :type gamma: float

        :arg default_radius: Default radius for atoms whose radii is not set
            as a keyword argument. Default is 7.5
        :type default_radius: float

        Keywords in keyword arguments must match those in *atom_identifiers*.
        Values of keyword arguments must be :class:`float`."""

        self._identifiers = identifiers
        radii = np.ones(len(identifiers)) * default_radius

        for i, identifier in enumerate(identifiers):
            radii[i] = kwargs.get(identifier, default_radius)
        self._radii = radii
        self._gamma = float(gamma)
        self._debug = bool(kwargs.get('debug', False))

    def getRadii(self):
        """Returns a copy of radii array."""

        return self._radii.copy()

    def getGamma(self):
        """Returns the uniform force constant value."""

        return self._gamma

    def gamma(self, dist2, i, j):
        """Returns force constant."""

        cutoff = (self._radii[i] + self._radii[j])
        cutoff2 = cutoff ** 2

        if dist2 < cutoff2:
            gamma = self._gamma
        else:
            gamma = 0
        if self._debug:
            print(' '.join([self._identifiers[i] + '_' + str(i), '--',
                  self._identifiers[j] + '_' + str(j),
                  'effective cutoff:', str(cutoff), 'distance:',
                  str(dist2**0.5), 'gamma:', str(gamma)]))  # PY3K: OK
        return gamma


class GammaED(Gamma):
    """Facilitate setting the spring constant based on 
    sequence distance and spatial distance as in ed-ENM [OL10]_.
    This ENM is refined based on comparison to essential dynamics 
    from MD simulations and can reproduce flexibility in NMR ensembles.

    It has also been implemented in FlexServ [CJ09]_ and 
    used in MDdMD [SP12]_ and GOdMD [SP13]_.

    The sequence distance-dependent term is Cseq/(S**2)
    for S=abs(i,j) <= Slim

    The structure distance-dependent term is (Ccart/dist)**Ex

    .. [OL10] Orellana L, Rueda M, Ferrer-Costa C, Lopez-Blanco JR, Chacón P, Orozco M. 
       Approaching Elastic Network Models to Molecular Dynamics Flexibility.
       *J Chem Theory Comput* **2010** 6(9):2910-23.

    .. [CJ09] Camps J, Carrillo O, Emperador A, Orellana L, Hospital A, Rueda M, 
       Cicin-Sain D, D'Abramo M, Gelpí JL, Orozco M.
       FlexServ: an integrated tool for the analysis of protein flexibility.
       *Bioinformatics* **2009** 25(13):1709-10.

    .. [SP12] Sfriso P, Emperador A, Orellana L, Hospital A, Gelpí JL, Orozco M.
       Finding Conformational Transition Pathways from Discrete Molecular Dynamics Simulations.
       *J Chem Theory Comput* **2012** 8(11):4707-18.

    .. [SP13] Sfriso P, Hospital A, Emperador A, Orozco M. 
       Exploration of conformational transition pathways from coarse-grained simulations.
       *Bioinformatics* **2013** 29(16):1980-6.     
       
    **Example**:

    Let's parse coordinates from a PDB file.

    .. ipython:: python

       from prody import *
       ubi = parsePDB('1aar', chain='A', subset='calpha')

    In the above we parsed only the atoms needed for this calculation, i.e.
    Cα atoms from chain A.

    We build the Hessian matrix using GOdMD spring constants as
    follows;

    .. ipython:: python

       gamma = GammaGOdMD(ubi)
       anm = ANM('')
       anm.buildHessian(ubi, gamma=gamma)
    
    """

    def __init__(self, atoms, Ccart=6., Ex=6, Cseq=60., Slim=3):
        """Setup the parameters.

        :arg atoms: A set of atoms
        :type atoms: :class:`.Atomic`

        :arg Ccart: Multiplication constant inside the exponential
            in the spatial distance-dependent term 
            Default is 6.
        :type Ccart: float

        :arg Ex: Exponent in spatial distance-dependent term
            Default is 6
        :type sheet: float

        :arg Cseq: Multiplication constant in sequence distance-dependent term 
            Default is 60.
        :type Cseq: float

        :arg Slim: Sequence distance limit for sequence distance-dependence
            This limit is used with a less-or-equal operator
            Default is 3
        :type Slim: float

    """

        if not isinstance(atoms, Atomic):
            raise TypeError('atoms must be an Atomic instance')

        n_atoms = atoms.numAtoms()
        if n_atoms < 3:
            raise ValueError('number of atoms must be larger than 2')

        rnum = atoms.getResindices()
        assert rnum is not None, 'residue numbers must be set'

        Ccart = float(Ccart)
        assert Ccart > 0, 'gamma must be greater than 0'

        Cseq = float(Cseq)
        assert Cseq > 0, 'Cseq must be greater than 0'

        self._Ccart = Ccart
        self._Ex = Ex

        self._Cseq = Cseq
        self._Slim = Slim

    def gamma(self, dist2, i, j):
        """Returns force constant."""

        Ccart = self._Ccart
        Ex = self._Ex

        Cseq = self._Cseq
        Slim = self._Slim

        S = abs(i-j)
        if S <= Slim:
            return Cseq/(S**2)
        else:
            dist = dist2**0.5
            return (Ccart/dist)**Ex

GammaGOdMD = GammaED



"""This module defines utility functions for solving eigenvalues."""

import numpy as np
from .misctools import importLA, div0
from .logger import LOGGER

__all__ = ['solveEig', 'ZERO']

ZERO = 1e-6

def solveEig(M, n_modes=None, zeros=False, turbo=True, expct_n_zeros=None, reverse=False, **kwargs):
    linalg = importLA()
    dof = M.shape[0]

    nproc = kwargs.get('nproc', 0)

    if expct_n_zeros is None:
        expct_n_zeros = 0
        warn_zeros = False
    else:
        warn_zeros = True

    if n_modes is None:
        eigvals = None
        n_modes = dof
    else:
        if n_modes >= dof:
            eigvals = None
            n_modes = dof
        else:
            if reverse:
                eigvals = (dof-expct_n_zeros-n_modes, dof-1)
            else:
                eigvals = (0, n_modes+expct_n_zeros-1)

    def _eigh(M, eigvals=None, turbo=True):
        if linalg.__package__.startswith('scipy'):
            from scipy.sparse import issparse

            if eigvals:
                turbo = False
            if not issparse(M):
                try:
                    values, vectors = linalg.eigh(M, turbo=turbo, eigvals=eigvals)
                except TypeError:
                    if turbo:
                        values, vectors = linalg.eigh(M, driver='evd', subset_by_index=eigvals)
                    else:
                        values, vectors = linalg.eigh(M, subset_by_index=eigvals)
            else:
                try:
                    from scipy.sparse import linalg as scipy_sparse_la
                except ImportError:
                    raise ImportError('failed to import scipy.sparse.linalg, '
                                      'which is required for sparse matrix '
                                      'decomposition')
                if eigvals:
                    j = eigvals[0]
                    k = eigvals[-1] + 1
                else:
                    j = 0
                    k = dof

                if k >= dof:
                    k -= 1
                    LOGGER.warning('Cannot calculate all eigenvalues for sparse matrices, thus '
                                   'the last eigenvalue is omitted. See scipy.sparse.linalg.eigsh '
                                   'for more information')
                values, vectors = scipy_sparse_la.eigsh(M, k=k, which='SA')
                values = values[j:k]
                vectors = vectors[:, j:k]
        else:
            if n_modes is not None:
                LOGGER.info('Scipy is not found, all modes were calculated.')
            else:
                n_modes = dof
            values, vectors = linalg.eigh(M)
        return values, vectors

    def _calc_n_zero_modes(M):
        from scipy.sparse import issparse

        if not issparse(M):
            w = linalg.eigvalsh(M)
        else:
            try:
                from scipy.sparse import linalg as scipy_sparse_la
            except ImportError:
                raise ImportError('failed to import scipy.sparse.linalg, '
                                    'which is required for sparse matrix '
                                    'decomposition')
            w, _ = scipy_sparse_la.eigsh(M, k=dof-1, which='SA')
        n_zeros = sum(w < ZERO)
        return n_zeros

    if nproc > 0:
        try:
            from threadpoolctl import threadpool_limits
        except ImportError:
            raise ImportError('Please install threadpoolctl to control threads')

        with threadpool_limits(limits=nproc, user_api="blas"):
            values, vectors = _eigh(M, eigvals, turbo)
    else:
        values, vectors = _eigh(M, eigvals, turbo)
    n_zeros = sum(values < ZERO)

    if warn_zeros:
        if n_zeros < n_modes + expct_n_zeros:
            if n_zeros < expct_n_zeros:
                LOGGER.warning('Fewer than %d (%d) zero eigenvalues were calculated.'%(expct_n_zeros, n_zeros))
            elif n_zeros > expct_n_zeros:
                LOGGER.warning('More than %d (%d) zero eigenvalues were calculated.'%(expct_n_zeros, n_zeros))
        else:
            LOGGER.warning('More than %d zero eigenvalues were detected.'%expct_n_zeros)

    if not zeros:
        if n_zeros > expct_n_zeros:
            if n_zeros == n_modes + expct_n_zeros and n_modes < dof:
                LOGGER.debug('Determing the number of zero eigenvalues...')
                # find the actual number of zero modes
                n_zeros = _calc_n_zero_modes(M)
                LOGGER.debug('%d zero eigenvalues detected.'%n_zeros)
            LOGGER.debug('Solving for additional eigenvalues...')

            if n_modes < dof:
                start = min(n_modes+expct_n_zeros, dof-1); end = min(n_modes+n_zeros-1, dof-1)
                values_, vectors_ = _eigh(M, eigvals=(start, end))
                values = np.concatenate((values, values_))
                vectors = np.hstack((vectors, vectors_))

        # final_n_modes may exceed len(eigvals) - no need to fix for the sake of the simplicity of the code
        final_n_modes = n_zeros + n_modes
        eigvals = values[n_zeros:final_n_modes]
        eigvecs = vectors[:, n_zeros:final_n_modes]
        invvals = 1 / eigvals
    else:
        eigvals = values[:n_modes]
        eigvecs = vectors[:, :n_modes]
        invvals = div0(1, values)
        invvals[:n_zeros] = 0.
        invvals = invvals[:n_modes]

    if reverse:
        invvals = invvals[::-1]
        eigvals = eigvals[::-1]
        eigvecs = eigvecs[:, ::-1]

    return eigvals, eigvecs, invvals



import numpy as np

from prody.chromatin.functions import div0
from prody.utilities import importLA
from prody import LOGGER

__all__ = ['VCnorm', 'SQRTVCnorm', 'Filenorm', 'SCN']

def VCnorm(M, **kwargs):
    """ Performs vanilla coverage normalization on matrix *M*."""

    total_count = kwargs.get('total_count', 'original')

    C = np.diag(div0(1., np.sum(M, axis=0)))
    R = np.diag(div0(1., np.sum(M, axis=1)))

    # N = R * M * C
    N = np.dot(np.dot(R,M),C)

    if total_count == 'original':
        total_count = np.sum(M)

    if total_count is not None:
        sum_N = np.sum(N)
        k = total_count / sum_N
        N = N * k
    return N

def SQRTVCnorm(M, **kwargs):
    """ Performs square-root vanilla coverage normalization on matrix *M*."""

    total_count = kwargs.get('total_count', 'original')

    C = np.diag(np.sqrt(div0(1., np.sum(M, axis=0))))
    R = np.diag(np.sqrt(div0(1., np.sum(M, axis=1))))

    # N = R * M * C
    N = np.dot(np.dot(R,M),C)

    if total_count == 'original':
        total_count = np.sum(M)

    if total_count is not None:
        sum_N = np.sum(N)
        k = total_count / sum_N
        N = N * k
    return N

def SCN(M, **kwargs):
    """ Performs Sequential Component Normalization on matrix *M*.
    
    .. [AC12] Cournac A, Marie-Nelly H, Marbouty M, Koszul R, Mozziconacci J. 
       Normalization of a chromosomal contact map. *BMC Genomics* **2012**. 
    """

    total_count = kwargs.pop('total_count', None)
    max_loops = kwargs.pop('max_loops', 100)
    tol = kwargs.pop('tol', 1e-5)

    N = M.copy()
    n = 0
    d0 = None
    p = 1
    last_p = None

    while True:
        C = np.diag(div0(1., np.sum(N, axis=0)))
        N = np.dot(N, C)

        R = np.diag(div0(1., np.sum(N, axis=1)))
        N = np.dot(R, N)

        n += 1

        # check convergence of symmetry
        d = np.mean(np.abs(N - N.T))
        
        if d0 is not None:
            p = div0(d, d0)
            dp = np.abs(p - last_p)
            if dp < tol:
                break
        else:
            d0 = d
        LOGGER.debug('Iteration {0}: d = {1}, p = {2}'.format(str(n), str(d), str(p)))
        last_p = p
        
        if max_loops is not None:
            if n >= max_loops:
                LOGGER.warn('The SCN algorithm did not converge after {0} '
                            'iterations.'.format(max_loops))
                break
    # guarantee symmetry
    N = (N + N.T) / 2.
    if total_count == 'original':
        total_count = np.sum(M)

    if total_count is not None:
        sum_N = np.sum(N)
        k = total_count / sum_N
        N = N * k
    return N

def bnewt(A, mask=[], tol = 1e-6, delta_lower = 0.1, delta_upper = 3, fl = 0, check = 1, largemem = 0, chunk_size = 10000):
    """
    BNEWT A balancing algorithm for symmetric matrices
    X = BNEWT(A) attempts to find a vector X such that
    diag(X)*A*diag(X) is close to doubly stochastic. A must
    be symmetric and nonnegative.
    
    X0: initial guess. TOL: error tolerance.
    delta/Delta: how close/far balancing vectors can get
    to/from the edge of the positive cone.
    We use a relative measure on the size of elements.
    FL: intermediate convergence statistics on/off.
    RES: residual error, measured by norm(diag(x)*A*x - e).
    """
    # see details in Knight and Ruiz (2012)
    (n,m) = A.shape
    #np.seterr(divide='ignore')
    print('Verifying Matrix')
    if (n != m):
        print('Matrix must be symmetric to converge')
        return 'NaN'
    if (check):
        for i in range(0,n):
            for j in range(i,n):
                if (A[i][j] != A[j][i])or(A[i][j] < 0):
                    print('Matrix must be symmetric and nonnegative to converge')
                    return 'NaN'
        print('Check OK\n')
    else:
        print('Check escaped\n')
  
    e        = np.ones((n,1))
    e[mask]  = 0
    #res      = np.empty((n,1))
    
    g        = 0.9
    etamax   = 0.1
    eta      = etamax
    stop_tol = tol*0.5
    x        = e #initial guess
    rt       = tol*tol
    if largemem:
        v      = x * chunking_dot(A,x,chunk_size=chunk_size)
    else:
        v      = x*np.dot(A,x)
    rk       = 1 - v
    rk[mask] = 0
    rho_km1  = np.dot(np.transpose(rk),rk)
    rout     = rho_km1
    rold     = rout
    
    MVP = 0 #matrix vector products
    i = 0
  
    while rout > rt:
        i = i+1
        k=0
        y=e
        innertol = max(eta*eta*rout,rt)
    
        while rho_km1 > innertol: #inner iteration by CG
            k = k+1
            if k==1:
                with np.errstate(invalid='ignore'):
                    Z       = rk/v
                Z[mask] = 0
                p       = Z
                rho_km1 = np.dot(np.transpose(rk),Z)
            else:
                beta = rho_km1/rho_km2
                p    =  Z + beta*p
      
            #update search direction 
            if largemem:
                w   = x*chunking_dot(A,x*p,chunk_size=chunk_size) + v*p
            else:
                w   = x*np.dot(A,x*p) + v*p
      
            alpha = rho_km1/np.dot(np.transpose(p),w)
            ap = alpha*p
      
            #test distance to boundary of cone
            ynew = y + ap
            if min(np.delete(ynew,mask)) <= delta_lower:
                if delta_lower == 0:
                    break
                else:
                    ind = np.nonzero(ap < 0)
                    gamma = min((delta_lower - y[ind])/ap[ind])
                    y = y + gamma*ap
                    break
                if max(ynew) >= delta_upper:
                    ind = np.nonzero(ynew > delta_upper)
                    gamma = min((delta_upper-y[ind])/ap[ind])
                    y = y + gamma*ap
                    break
      
            y       = ynew
            rk      = rk - alpha*w
            rho_km2 = rho_km1
            with np.errstate(invalid='ignore'):
                Z       = rk/v
            Z[mask] = 0
            rho_km1 = np.dot(np.transpose(rk),Z)
        #end inner iteration
    
        x        = x*y
        if largemem:
            v      = x * chunking_dot(A,x,chunk_size=chunk_size)
        else:
            v      = x*np.dot(A,x)
        rk       = 1-v
        rk[mask] = 0
        rho_km1  = np.dot(np.transpose(rk),rk)
        rout     = rho_km1
        MVP      = MVP + k + 1
        #print MVP,res
        #update inner iteration stopping criterion
        rat      = rout/rold
        rold     = rout
        res_norm = math.sqrt(rout)
        eta_o    = eta
        eta      = g*rat
    
        if g*eta_o*eta_o > 0.1:
            eta = max(eta,g*eta_o*eta_o)
        eta = max(min(eta,etamax),stop_tol/res_norm)
    
        if fl == 1:
            print('%3d %6d %.3f \n' % (i,k,r_norm))
      
        if MVP > 50000:
            break
    #end outer
  
    print('Matrix vector products = %6d' % MVP)
    #x = np.array(x)
    #x[mask] = 0
    return x

def KRnorm(M, mask=None, **kwargs):
    """
    Uses Knight-Ruiz normalization to balance the matrix.
    """

    x = bnewt(M, mask=mask, check=0, **kwargs)*100
    M *= X 
    M *= X.T

    return M

def Filenorm(M, **kwargs):
    """ Performs normalization on matrix *M* given a file. *filename* specifies 
    the path to the file. The file should be one-column, and ideally has the 
    same number of entries with the size of *M* (extra entries will be ignored). 
    Say *F* is vector of the normalization factors, *N* is the normalized matrix, 
    if *expected* is **True**, ``N[i,j] = M[i,j]/F[i]/F[j]``. If *expected* is
    **True**, ``N[i,j] = M[i,j]/F[|i-j|]``."""

    filename = kwargs.get('filename')
    expected = kwargs.get('expected', False)

    if filename is None:
        raise IOError("'filename' is not specified.")
    factors = np.loadtxt(filename)
    L = M.shape[0]
    if not expected:
        factors.resize(L)
        norm_mat = np.outer(factors, factors)
        
        N = div0(M, norm_mat)
        return N
    else:
        N = np.zeros(M.shape)
        for i in range(L):
            for j in range(L):
                N[i,j] = div0(M[i,j], factors[abs(i-j)])
        return N



# -*- coding: utf-8 -*-
"""This module defines functions for performing adaptive ANM."""

from prody.atomic import Atomic, AtomMap
import time
from numbers import Integral, Number
import numpy as np

from prody import LOGGER
from prody.utilities import getCoords, importLA
from prody.measure import calcRMSD, calcDistance, superpose
from prody.ensemble import Ensemble

from .functions import calcENM
from .modeset import ModeSet

__all__ = ['calcAdaptiveANM', 'AANM_ONEWAY', 'AANM_ALTERNATING', 'AANM_BOTHWAYS', 'AANM_DEFAULT']

AANM_ALTERNATING = 0
AANM_ONEWAY = 1
AANM_BOTHWAYS = 2

AANM_DEFAULT = AANM_ALTERNATING

norm = importLA().norm

def checkInput(a, b, **kwargs):
    coordsA = getCoords(a)
    if isinstance(a, Atomic):
        title = a.getTitle()
        atoms = a
    else:
        title = None
        atoms = None

    coordsB = getCoords(b)
    
    if title is None:
        if isinstance(b, Atomic):
            title = b.getTitle()
            atoms = b
        else:
            title = 'Unknown'
            atoms = None

    maskA = a.getFlags("mapped") if isinstance(a, AtomMap) else 1.
    maskB = b.getFlags("mapped") if isinstance(b, AtomMap) else 1.
    weights = maskA * maskB

    if np.isscalar(weights):
        weights = None
    
    if np.isscalar(maskA):
        maskA = None

    if np.isscalar(maskB):
        maskB = None

    aligned = kwargs.get('aligned', False)
    if not aligned:
        coordsA, _ = superpose(coordsA, coordsB, weights)

    rmsd = calcRMSD(coordsA, coordsB, weights)
    LOGGER.info('Initialized Adaptive ANM with RMSD {:4.3f}\n'.format(rmsd))

    return coordsA, coordsB, title, atoms, weights, maskA, maskB, rmsd

def getTitle(structure, def_title='structure'):
    if isinstance(structure, Atomic):
        title = structure.getTitle()
    else:
        title = def_title

    return title

def calcStep(initial, target, n_modes, ensemble, defvecs, rmsds, mask=None, callback_func=None, **kwargs):
    """Runs a single step of adaptive ANM. 
    Modes will be calculated for *initial* with a square cumulative overlap above a threshold defined by 
    *Fmin* and used for transitioning towards *target*.
    """

    Fmin = kwargs.get('Fmin', None)
    f = kwargs.get('f', 0.2)

    Fmin_max = kwargs.get('Fmin_max', 0.6)
    resetFmin = kwargs.get('resetFmin', False)

    weights = ensemble.getWeights()
    if weights is not None:
        weights = weights.flatten()
    #coords_init, _ = superpose(initial, target, weights) # we should keep this off otherwise RMSD calculations are off
    coords_init = initial
    coords_tar = target

    dof = coords_init.shape[0] - 6
    n_max_modes = kwargs.get('n_max_modes', None)
    if n_max_modes is None:
        n_max_modes = dof

    if n_max_modes < 1:
        n_max_modes = int(n_max_modes * dof)
    if n_max_modes > dof:
        n_max_modes = dof

    if n_modes > n_max_modes:
        n_modes = n_max_modes

    model = kwargs.pop('model', 'anm')
    anm, _ = calcENM(coords_init, select=mask, mask=mask, 
                     model=model, trim='trim', n_modes=n_modes, 
                     **kwargs)

    if mask is not None:
        anm.masked = False

    defvec = coords_tar - coords_init
    d = defvec.flatten()
    if weights is not None:
        d *= weights.repeat(3)
    defvecs.append(d)

    if Fmin is None:
        if resetFmin:
            Fmin = 0.  # Select the first mode only
        else:
            Fmin = 1 - np.sqrt(norm(defvecs[-1])/norm(defvecs[0]))

    if Fmin > Fmin_max:
        Fmin = Fmin_max

    overlaps = np.dot(d, anm.getEigvecs())

    normalised_overlaps = overlaps / norm(d)
    c_sq = np.cumsum(np.power(normalised_overlaps, 2), axis=0)

    if Fmin == 0 and resetFmin:
        torf_Fmin = np.zeros(c_sq.shape, dtype=bool)
        argmax_overlap = np.argmax(abs(normalised_overlaps))
        torf_Fmin[argmax_overlap] = True
    else:
        torf_Fmin = c_sq <= Fmin
        if np.any(torf_Fmin) and not np.all(torf_Fmin):
            i = np.where(torf_Fmin)[0].max()
            torf_Fmin[i+1] = True

        if not np.any(torf_Fmin):
            torf_Fmin[0] = True

    selected_mode_indices = np.arange(anm.numModes())[torf_Fmin]

    n_sel_modes = len(selected_mode_indices)

    modes = ModeSet(anm, selected_mode_indices)
    c_sq_crit = c_sq[torf_Fmin].max()

    if n_sel_modes == 1:
        LOGGER.info('Using 1 mode with square overlap {0}'
                    .format('%4.3f'%c_sq_crit))
    else:
        LOGGER.info('Using {0} modes with square cumulative overlap {1}'
                    .format(n_sel_modes, '%4.3f'%c_sq_crit))

    if n_sel_modes > n_modes-5:
        n_modes *= 2

    if n_modes > dof:
        n_modes = dof

    v = modes.getEigvecs().dot(overlaps[torf_Fmin])
    s = f * v.dot(d) / v.dot(v)

    # update coords_init
    coords_init += s * v.reshape(coords_init.shape)
    # initial[:] = coords_init[:] # turn this on in case coords_init is not initial in the future
    rmsd = calcRMSD(coords_init, coords_tar, weights)
    rmsds.append(rmsd)

    if callback_func is not None:
        cbkwargs = {'init': coords_init, 
                    'tar': coords_tar, 
                    'modes': modes, 
                    'defvec': d,
                    'c_sq': c_sq_crit,
                    'rmsd': rmsd}
        callback_func(**cbkwargs)

    # deposit 
    ensemble.addCoordset(coords_init.copy())
    converged = checkConvergence(rmsds, coords_init, **kwargs)

    if converged:
        n_modes = 0

    LOGGER.info('Current RMSD is {:4.3f}\n'.format(rmsd))

    return n_modes

def checkConvergence(rmsds, coords, **kwargs):
    """Check convergence of adaptive ANM. 

    Convergence is reached if one of three conditions is met:
    1. Difference between *rmsds* from previous step to current < *min_rmsd_diff*
    2. Current rmsd < *target_rmsd* for the last five runs
    3. A node in *coords* gets disconnected from another by > *cutoff*
    """
    min_rmsd_diff = kwargs.get('min_rmsd_diff', 0.05)
    target_rmsd = kwargs.get('target_rmsd', 1.0)
    cutoff = kwargs.get('cutoff', 15)

    if len(rmsds) > 4:
        drmsd = np.abs(np.diff(rmsds))

        if np.all(drmsd[-4:] < min_rmsd_diff):
            LOGGER.warn(
                'The RMSD decrease fell below {0}'.format(min_rmsd_diff))
            return True

    if rmsds[-1] < target_rmsd:
        LOGGER.warn('The RMSD fell below target RMSD {0}'.format(target_rmsd))
        return True

    if checkDisconnection(coords, cutoff):
        LOGGER.warn('Disconnections were found in one of the structures {0}')
        return True

    return False


def checkDisconnection(coords, cutoff):
    """Check disconnection of ANM, i.e. a node in *coords* gets 
    disconnected from another by > *cutoff*. This is one of the 
    stopping criteria for adaptive ANM. 
    """

    all_dists = np.array([calcDistance(coords, entry) for entry in coords])
    min_dists = np.array([np.min([np.min(all_dists[i, :i]), np.min(all_dists[i, i+1:])])
                          for i in range(1, coords.shape[0]-1)])
    if max(min_dists) > cutoff:
        LOGGER.warn('A bead has become disconnected. '
                    'Adaptive ANM cannot proceed without unrealistic deformations')
        return True

    return False

def calcAdaptiveANM(a, b, n_steps, mode=AANM_DEFAULT, **kwargs):
    """Runs adaptive ANM analysis of proteins ([ZY09]_) that creates a path that 
    connects two conformations using normal modes.

    This function can be run in three modes:
    
    1. *AANM_ONEWAY*: all steps are run in one direction: from *a* to *b*.

    2. *AANM_ALTERNATING*: steps are run in alternating directions: from *a* to *b*, 
        then *b* to *a*, then back again, and so on.

    3. *AANM_BOTHWAYS*: steps are run in one direction (from *a* to 
        *b*) until convergence is reached and then the other way.

    This also implementation differs from the original one in that it sorts the 
    modes by overlap prior to cumulative overlap calculations for efficiency.

    .. [ZY09] Zheng Yang, Peter Májek, Ivet Bahar. Allosteric Transitions of 
            Supramolecular Systems Explored by Network Models: Application to 
            Chaperonin GroEL. *PLOS Comp Biol* **2009** 40:512-524.

    :arg a: structure A for the transition
    :type a: :class:`.Atomic`, :class:`~numpy.ndarray`

    :arg b: structure B for the transition
    :type b: :class:`.Atomic`, :class:`~numpy.ndarray`

    :arg n_steps: the maximum number of steps to be calculated. For *AANM_BOTHWAYS*, 
        this means the maximum number of steps from each direction
    :type n_steps: int

    :arg mode: the way of the calculation to be performed, which can be either *AANM_ONEWAY*, 
        *AANM_ALTERNATING*, or *AANM_BOTHWAYS*. Default is *AANM_ALTERNATING*
    :type mode: int

    :kwarg f: step size. Default is 0.2
    :type f: float

    :kwarg Fmin: cutoff for selecting modes based on square cumulative overlaps
        Default is **None**, which automatically determines and adapts *Fmin* on the fly.
    :type Fmin: float

    :kwarg Fmin_max: maximum value for *Fmin* when it is automatically determined
        Default is 0.6
    :type Fmin_max: float

    :arg min_rmsd_diff: cutoff for rmsds converging. Default is 0.05
    :type min_rmsd_diff: float

    :kwarg target_rmsd: target rmsd for stopping. Default is 1.0
    :type target_rmsd: float

    :kwarg n_modes: the number of modes to be calculated for the first run. *n_modes* 
        will be dynamically adjusted later as the calculation progresses. Default is 20
    :type n_modes: int

    :kwarg n_max_modes: the maximum number of modes to be calculated in each run. 
        Default is **None**, which allows as many as degree of freedom
    :type n_max_modes: int

    :kwarg callback_func: a callback function that can be used to collect quantities 
        from each iteration. The function must accept `**kwargs` as its only input. 
        Keywords in `kwargs` are:
        'init': the initial coordinate; 
        'tar': the target coordinate; 
        'modes': a :class:`.ModeSet` of selected modes; 
        'defvec': the deformation vector; 
        'c_sq': the critical square cumulative overlap; 
        'rmsd': the RMSD between the two structures after the deformation.
    :type callback_func: func

    Please see keyword arguments for calculating the modes in :func:`.calcENM`.
    """

    if mode == AANM_ONEWAY:
        return calcOneWayAdaptiveANM(a, b, n_steps, **kwargs)
    elif mode == AANM_ALTERNATING:
        return calcAlternatingAdaptiveANM(a, b, n_steps, **kwargs)
    elif mode == AANM_BOTHWAYS:
        return calcBothWaysAdaptiveANM(a, b, n_steps, **kwargs)
    else:
        raise ValueError('unknown aANM mode: %d'%mode)

def calcOneWayAdaptiveANM(a, b, n_steps, **kwargs):
    """Runs one-way adaptivate ANM. """

    n_modes = kwargs.pop('n_modes', 20)

    coordsA, coordsB, title, atoms, weights, maskA, maskB, rmsd = checkInput(a, b, **kwargs)
    coordsA = coordsA.copy()

    LOGGER.timeit('_prody_calcAdaptiveANM')
    n = 0
    resetFmin = True
    defvecs = []
    rmsds = [rmsd]
    ensemble = Ensemble(title + '_aANM')
    ensemble.setAtoms(atoms)
    ensemble.setCoords(coordsB)
    ensemble.setWeights(weights)
    ensemble.addCoordset(coordsA.copy())
    while n < n_steps:
        LOGGER.info('\nStarting cycle {0} with initial structure {1}'.format(n+1, title))                                                            
        n_modes = calcStep(coordsA, coordsB, n_modes, ensemble, defvecs, rmsds, mask=maskA,
                           resetFmin=resetFmin, **kwargs)
        n += 1
        resetFmin = False
        if n_modes == 0:
            LOGGER.report('One-way Adaptive ANM converged in %.2fs.', '_prody_calcAdaptiveANM')
            break

    return ensemble


def calcAlternatingAdaptiveANM(a, b, n_steps, **kwargs):
    """Runs alternating adaptivate ANM. """

    n_modes = kwargs.pop('n_modes', 20)

    coordsA, coordsB, title, atoms, weights, maskA, maskB, rmsd = checkInput(a, b, **kwargs)
    coordsA = coordsA.copy()
    coordsB = coordsB.copy()

    LOGGER.timeit('_prody_calcAdaptiveANM')
    n = 0
    resetFmin = True
    defvecs = []
    rmsds = [rmsd]
    ensA = Ensemble('A')
    ensA.setCoords(coordsA)
    ensA.setWeights(weights)
    ensA.addCoordset(coordsA.copy())

    ensB = Ensemble('B')
    ensB.setCoords(coordsB.copy())
    ensB.setWeights(weights)
    ensB.addCoordset(coordsB.copy())

    while n < n_steps:
        LOGGER.info('\nStarting cycle {0} with {1}'.format(n + 1, getTitle(a, 'structure A')))
        n_modes = calcStep(coordsA, coordsB, n_modes, ensA, defvecs, rmsds, mask=maskA,
                           resetFmin=resetFmin, **kwargs)
        resetFmin = False

        if n_modes == 0:
            LOGGER.report('Alternating Adaptive ANM converged in %.2fs.', '_prody_calcAdaptiveANM')
            break

        LOGGER.info('\nContinuing cycle {0} with structure {1}'.format(n+1, getTitle(b, 'structure B')))
        n_modes = calcStep(coordsB, coordsA, n_modes, ensB, defvecs, rmsds, mask=maskB,
                           resetFmin=resetFmin, **kwargs)
        n += 1

        if n_modes == 0:
            LOGGER.report('Alternating Adaptive ANM converged in %.2fs.', '_prody_calcAdaptiveANM')
            break
    
    ensemble = ensA + ensB[::-1]
    ensemble.setTitle(title + '_aANM')
    ensemble.setAtoms(atoms)
    ensemble.setCoords(ensB.getCoords())

    return ensemble


def calcBothWaysAdaptiveANM(a, b, n_steps, **kwargs):
    """Runs both-way adaptivate ANM. """

    n_modes0 = n_modes = kwargs.pop('n_modes', 20)

    coordsA, coordsB, title, atoms, weights, maskA, maskB, rmsd = checkInput(a, b, **kwargs)
    coordsA = coordsA.copy()
    coordsB = coordsB.copy()

    LOGGER.timeit('_prody_calcAdaptiveANM')
    n = 0
    resetFmin = True
    defvecs = []
    rmsds = [rmsd]
    ensA = Ensemble('A')
    ensA.setCoords(coordsA)
    ensA.setWeights(weights)
    ensA.addCoordset(coordsA.copy())

    ensB = Ensemble('B')
    ensB.setCoords(coordsB.copy())
    ensB.setWeights(weights)
    ensB.addCoordset(coordsB.copy())
    
    while n < n_steps:
        LOGGER.info('\nStarting cycle {0} with {1}'.format(n + 1, getTitle(a, 'structure A')))
        n_modes = calcStep(coordsA, coordsB, n_modes, ensA, defvecs, rmsds, mask=maskA,
                           resetFmin=resetFmin, **kwargs)
        n += 1
        resetFmin = False

        if n_modes == 0:
            break
    
    n = 0
    n_modes = n_modes0
    resetFmin = True
    while n < n_steps:
        LOGGER.info('\nStarting cycle {0} with structure {1}'.format(n+1, getTitle(b, 'structure B')))
        n_modes = calcStep(coordsB, coordsA, n_modes, ensB, defvecs, rmsds, mask=maskB,
                           resetFmin=resetFmin, **kwargs)
        n += 1
        resetFmin = False

        if n_modes == 0:
            LOGGER.report('Alternating Adaptive ANM converged in %.2fs.', '_prody_calcAdaptiveANM')
            break

    ensemble = ensA + ensB[::-1]
    ensemble.setTitle(title + '_aANM')
    ensemble.setAtoms(atoms)
    ensemble.setCoords(ensB.getCoords())

    LOGGER.report('Both-way Adaptive ANM converged in %.2fs.', '_prody_calcAdaptiveANM')

    return ensemble

